from __future__ import annotations

import contextlib
import importlib
import importlib.metadata
import itertools
import operator
from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING, Any

import _pytest
import numpy as np
import pandas as pd
import pytest
from packaging.requirements import Requirement
from packaging.version import parse as vparse

import ibis
import ibis.common.exceptions as com
from ibis import util
from ibis.backends.base import CanCreateDatabase, CanCreateSchema, _get_backend_names
from ibis.conftest import WINDOWS
from ibis.util import promote_tuple

if TYPE_CHECKING:
    from ibis.backends.tests.base import BackendTest


def compare_versions(module_name, given_version, op):
    try:
        current_version = importlib.metadata.version(module_name)
        return op(vparse(current_version), vparse(given_version))
    except importlib.metadata.PackageNotFoundError:
        return False


def is_newer_than(module_name, given_version):
    return compare_versions(module_name, given_version, operator.gt)


def is_older_than(module_name, given_version):
    return compare_versions(module_name, given_version, operator.lt)


TEST_TABLES = {
    "functional_alltypes": ibis.schema(
        {
            "id": "int32",
            "bool_col": "boolean",
            "tinyint_col": "int8",
            "smallint_col": "int16",
            "int_col": "int32",
            "bigint_col": "int64",
            "float_col": "float32",
            "double_col": "float64",
            "date_string_col": "string",
            "string_col": "string",
            "timestamp_col": "timestamp",
            "year": "int32",
            "month": "int32",
        }
    ),
    "diamonds": ibis.schema(
        {
            "carat": "float64",
            "cut": "string",
            "color": "string",
            "clarity": "string",
            "depth": "float64",
            "table": "float64",
            "price": "int64",
            "x": "float64",
            "y": "float64",
            "z": "float64",
        }
    ),
    "batting": ibis.schema(
        {
            "playerID": "string",
            "yearID": "int64",
            "stint": "int64",
            "teamID": "string",
            "lgID": "string",
            "G": "int64",
            "AB": "int64",
            "R": "int64",
            "H": "int64",
            "X2B": "int64",
            "X3B": "int64",
            "HR": "int64",
            "RBI": "int64",
            "SB": "int64",
            "CS": "int64",
            "BB": "int64",
            "SO": "int64",
            "IBB": "int64",
            "HBP": "int64",
            "SH": "int64",
            "SF": "int64",
            "GIDP": "int64",
        }
    ),
    "awards_players": ibis.schema(
        {
            "playerID": "string",
            "awardID": "string",
            "yearID": "int64",
            "lgID": "string",
            "tie": "string",
            "notes": "string",
        }
    ),
    "astronauts": ibis.schema(
        {
            "id": "int64",
            "number": "int64",
            "nationwide_number": "int64",
            "name": "string",
            "original_name": "string",
            "sex": "string",
            "year_of_birth": "int64",
            "nationality": "string",
            "military_civilian": "string",
            "selection": "string",
            "year_of_selection": "int64",
            "mission_number": "int64",
            "total_number_of_missions": "int64",
            "occupation": "string",
            "year_of_mission": "int64",
            "mission_title": "string",
            "ascend_shuttle": "string",
            "in_orbit": "string",
            "descend_shuttle": "string",
            "hours_mission": "float64",
            "total_hrs_sum": "float64",
            "field21": "int64",
            "eva_hrs_mission": "float64",
            "total_eva_hrs": "float64",
        }
    ),
}

# We want to check for exceptions in xfail tests for two reasons:
# * xfail tests without exception checking may hide problems.
#   For example, the implementation may work, but we may have an error in the test that needs to be fixed.
# * facilitates code reviews
# For now, many of our tests don't do this, and we're working to change this situation
# by improving all tests file by file. All files that have already been improved are
# added to this list to prevent regression.
FIlES_WITH_STRICT_EXCEPTION_CHECK = [
    "ibis/backends/tests/test_api.py",
    "ibis/backends/tests/test_array.py",
    "ibis/backends/tests/test_aggregation.py",
    "ibis/backends/tests/test_binary.py",
    "ibis/backends/tests/test_numeric.py",
    "ibis/backends/tests/test_column.py",
    "ibis/backends/tests/test_string.py",
    "ibis/backends/tests/test_temporal.py",
    "ibis/backends/tests/test_uuid.py",
    "ibis/backends/tests/test_window.py",
]

ALL_BACKENDS = set(_get_backend_names())


@pytest.fixture(scope="session")
def data_dir() -> Path:
    """Return the test data directory.

    Returns
    -------
    Path
        Test data directory
    """
    root = Path(__file__).absolute().parents[2]

    return root / "ci" / "ibis-testing-data"


def _get_backend_conf(backend_str: str):
    """Convert a backend string to the test class for the backend."""
    conftest = importlib.import_module(f"ibis.backends.{backend_str}.tests.conftest")
    return conftest.TestConf


def _get_backend_from_parts(parts: tuple[str, ...]) -> str | None:
    """Return the backend part of a test file's path parts.

    Examples
    --------
    >>> _get_backend_from_parts(("/", "ibis", "backends", "sqlite", "tests"))
    "sqlite"
    """
    try:
        index = parts.index("backends")
    except ValueError:
        return None
    else:
        return parts[index + 1]


def pytest_ignore_collect(path, config):
    # get the backend path part
    #
    # path is a py.path.local object hence the conversion to Path first
    backend = _get_backend_from_parts(Path(path).parts)
    if backend is None or backend not in _get_backend_names():
        return False

    # we evaluate the marker early so that we don't trigger
    # an import of conftest files for the backend, which will
    # import the backend and thus require dependencies that may not
    # exist
    #
    # alternatives include littering library code with pytest.importorskips
    # and moving all imports close to their use site
    #
    # the latter isn't tenable for backends that use multiple dispatch
    # since the rules are executed at import time
    mark_expr = config.getoption("-m")
    # we can't let the empty string pass through, since `'' in s` is `True` for
    # any `s`; if no marker was passed don't ignore the collection of `path`
    if not mark_expr:
        return False
    expr = _pytest.mark.expression.Expression.compile(mark_expr)
    # we check the "backend" marker as well since if that's passed
    # any file matching a backend should be skipped
    keep = expr.evaluate(lambda s: s in (backend, "backend"))
    return not keep


def pytest_collection_modifyitems(session, config, items):
    all_backends = _get_backend_names()
    additional_markers = []

    try:
        import pyspark
    except ImportError:
        pyspark = None

    unrecognized_backends = set()
    for item in items:
        # Yell loudly if unrecognized backend in notimpl, notyet or never
        for name in ("notimpl", "notyet", "never"):
            for mark in item.iter_markers(name=name):
                if backend := set(util.promote_list(mark.args[0])).difference(
                    ALL_BACKENDS
                ):
                    unrecognized_backends.add(
                        f"""Unrecognize backend(s) {backend} passed to {name} marker in
{item.path}::{item.originalname}"""
                    )

        # add the backend marker to any tests are inside "ibis/backends"
        parts = item.path.parts
        backend = _get_backend_from_parts(parts)
        if backend is not None and backend in all_backends:
            item.add_marker(getattr(pytest.mark, backend))
            item.add_marker(pytest.mark.backend)
        elif "backends" not in parts and not tuple(
            itertools.chain(
                *(item.iter_markers(name=name) for name in all_backends),
                item.iter_markers(name="backend"),
            )
        ):
            # anything else is a "core" test and is run by default
            if not any(item.iter_markers(name="benchmark")):
                item.add_marker(pytest.mark.core)

        # skip or xfail pyspark tests that run afoul of our non-ancient stack
        for _ in item.iter_markers(name="pyspark"):
            if not isinstance(item, pytest.DoctestItem):
                additional_markers.append(
                    (
                        item,
                        [
                            pytest.mark.skipif(
                                pyspark is not None
                                and vparse(pyspark.__version__) < vparse("3.3.3")
                                and vparse(np.__version__) >= vparse("1.24"),
                                reason="PySpark doesn't support numpy >= 1.24",
                            ),
                        ],
                    )
                )

    if unrecognized_backends:
        raise pytest.PytestCollectionWarning("\n" + "\n".join(unrecognized_backends))

    for item, markers in additional_markers:
        for marker in markers:
            item.add_marker(marker)


@cache
def _get_backends_to_test(
    keep: tuple[str, ...] = (),
    discard: tuple[str, ...] = (),
) -> list[Any]:
    """Get a list of `TestConf` classes of the backends to test."""
    backends = _get_backend_names()

    if discard:
        backends = backends.difference(discard)

    if keep:
        backends = backends.intersection(keep)

    return [
        pytest.param(
            backend,
            marks=[getattr(pytest.mark, backend), pytest.mark.backend],
            id=backend,
        )
        for backend in sorted(backends)
    ]


def pytest_runtest_call(item):
    """Dynamically add various custom markers."""
    backend = [
        getattr(backend, "name", lambda backend=backend: backend)()
        for key, backend in item.funcargs.items()
        if key.endswith(("backend", "backend_name", "backend_no_data"))
    ]
    if len(backend) > 1:
        raise ValueError(
            f"test {item.originalname} was supplied with multiple backend "
            f"objects simultaneously. This is likely due to a leaky fixture."
            f"Backends passed: {(back.name() for back in backend)}"
        )
    if not backend:
        # Check item path to see if test is in backend-specific folder
        backend = set(_get_backend_names()).intersection(item.path.parts)

    if not backend:
        return

    backend = next(iter(backend))

    def _filter_none_from_raises(kwargs):
        # Filter out any None values from kwargs['raises']
        # to cover any missing backend error types as defined in ibis/backends/tests/errors.py
        if (raises := kwargs.get("raises")) is not None:
            raises = tuple(filter(None, promote_tuple(raises)))
            if raises:
                kwargs["raises"] = raises
            else:
                # if filtering removes all of the values of raises pop the
                # argument otherwise it gets passed as an empty tuple and this
                # messes up xfail
                kwargs.pop("raises")
        return kwargs

    # Ibis hasn't exposed existing functionality
    # This xfails so that you know when it starts to pass
    for marker in item.iter_markers(name="notimpl"):
        if backend in marker.args[0]:
            if (
                item.location[0] in FIlES_WITH_STRICT_EXCEPTION_CHECK
                and "raises" not in marker.kwargs.keys()
            ):
                raise ValueError("notimpl requires a raises")
            kwargs = marker.kwargs.copy()
            kwargs.setdefault("reason", f"Feature not yet exposed in {backend}")
            kwargs = _filter_none_from_raises(kwargs)
            item.add_marker(pytest.mark.xfail(**kwargs))

    # Functionality is unavailable upstream (but could be)
    # This xfails so that you know when it starts to pass
    for marker in item.iter_markers(name="notyet"):
        if backend in marker.args[0]:
            if (
                item.location[0] in FIlES_WITH_STRICT_EXCEPTION_CHECK
                and "raises" not in marker.kwargs.keys()
            ):
                raise ValueError("notyet requires a raises")

            kwargs = marker.kwargs.copy()
            kwargs.setdefault("reason", f"Feature not available upstream for {backend}")
            kwargs = _filter_none_from_raises(kwargs)
            item.add_marker(pytest.mark.xfail(**kwargs))

    for marker in item.iter_markers(name="never"):
        if backend in marker.args[0]:
            if "reason" not in marker.kwargs.keys():
                raise ValueError("never requires a reason")
            kwargs = marker.kwargs.copy()
            kwargs = _filter_none_from_raises(kwargs)
            item.add_marker(pytest.mark.xfail(**kwargs))

    # Something has been exposed as broken by a new test and it shouldn't be
    # imperative for a contributor to fix it just because they happened to
    # bring it to attention  -- USE SPARINGLY
    for marker in item.iter_markers(name="broken"):
        if backend in marker.args[0]:
            if (
                item.location[0] in FIlES_WITH_STRICT_EXCEPTION_CHECK
                and "raises" not in marker.kwargs.keys()
            ):
                raise ValueError("broken requires a raises")

            kwargs = marker.kwargs.copy()
            kwargs.setdefault("reason", f"Feature is failing on {backend}")
            kwargs = _filter_none_from_raises(kwargs)
            item.add_marker(pytest.mark.xfail(**kwargs))

    for marker in item.iter_markers(name="xfail_version"):
        kwargs = marker.kwargs.copy()
        kwargs = _filter_none_from_raises(kwargs)
        if backend not in kwargs:
            continue

        provided_reason = kwargs.pop("reason", None)
        specs = kwargs.pop(backend)
        failing_specs = []
        for spec in specs:
            req = Requirement(spec)
            if req.specifier.contains(importlib.import_module(req.name).__version__):
                failing_specs.append(spec)
        reason = f"{backend} backend test fails with {backend}{specs}"
        if provided_reason is not None:
            reason += f"; {provided_reason}"
        if failing_specs:
            item.add_marker(pytest.mark.xfail(reason=reason, **kwargs))


@pytest.fixture(params=_get_backends_to_test(), scope="session")
def backend(request, data_dir, tmp_path_factory, worker_id) -> BackendTest:
    """Return an instance of BackendTest, loaded with data."""

    cls = _get_backend_conf(request.param)
    return cls.load_data(data_dir, tmp_path_factory, worker_id)


@pytest.fixture(scope="session")
def con(backend):
    """Instance of a backend client."""
    return backend.connection


@pytest.fixture(params=_get_backends_to_test(), scope="session")
def backend_no_data(request, data_dir, tmp_path_factory, worker_id):
    """Return an instance of BackendTest, with no data loaded."""
    cls = _get_backend_conf(request.param)
    return cls(data_dir=data_dir, tmpdir=tmp_path_factory, worker_id=worker_id)


@pytest.fixture(scope="session")
def con_no_data(backend_no_data):
    """Return an Ibis backend instance, with no data loaded."""
    return backend_no_data.connection


@pytest.fixture(scope="session")
def con_create_database(con):
    if isinstance(con, CanCreateDatabase):
        return con
    else:
        pytest.skip(f"{con.name} backend cannot create databases")


@pytest.fixture(scope="session")
def con_create_schema(con):
    if isinstance(con, CanCreateSchema):
        return con
    else:
        pytest.skip(f"{con.name} backend cannot create schemas")


@pytest.fixture(scope="session")
def con_create_database_schema(con):
    if isinstance(con, CanCreateDatabase) and isinstance(con, CanCreateSchema):
        return con
    else:
        pytest.skip(f"{con.name} backend cannot create both database and schemas")


def _setup_backend(request, data_dir, tmp_path_factory, worker_id):
    if (backend := request.param) == "duckdb" and WINDOWS:
        pytest.xfail(
            "windows prevents two connections to the same duckdb file "
            "even in the same process"
        )
    else:
        cls = _get_backend_conf(backend)
        return cls.load_data(data_dir, tmp_path_factory, worker_id)


@pytest.fixture(
    params=_get_backends_to_test(discard=("dask", "pandas")),
    scope="session",
)
def ddl_backend(request, data_dir, tmp_path_factory, worker_id):
    """Set up the backends that are SQL-based."""
    return _setup_backend(request, data_dir, tmp_path_factory, worker_id)


@pytest.fixture(scope="session")
def ddl_con(ddl_backend):
    """Instance of Client, already connected to the db (if applies)."""
    return ddl_backend.connection


@pytest.fixture(
    params=_get_backends_to_test(keep=("dask", "pandas", "pyspark")),
    scope="session",
)
def udf_backend(request, data_dir, tmp_path_factory, worker_id):
    """Runs the UDF-supporting backends."""
    cls = _get_backend_conf(request.param)
    return cls.load_data(data_dir, tmp_path_factory, worker_id)


@pytest.fixture(scope="session")
def udf_con(udf_backend):
    """Instance of Client, already connected to the db (if applies)."""
    return udf_backend.connection


@pytest.fixture(scope="session")
def alltypes(backend):
    return backend.functional_alltypes


@pytest.fixture(scope="session")
def json_t(backend):
    return backend.json_t


@pytest.fixture(scope="session")
def struct(backend):
    return backend.struct


@pytest.fixture(scope="session")
def sorted_alltypes(alltypes):
    return alltypes.order_by("id")


@pytest.fixture(scope="session")
def udf_alltypes(udf_backend):
    return udf_backend.functional_alltypes


@pytest.fixture(scope="session")
def batting(backend):
    return backend.batting


@pytest.fixture(scope="session")
def awards_players(backend):
    return backend.awards_players


@pytest.fixture
def analytic_alltypes(alltypes):
    return alltypes


@pytest.fixture(scope="session")
def df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="session")
def struct_df(struct):
    return struct.execute()


@pytest.fixture(scope="session")
def udf_df(udf_alltypes):
    return udf_alltypes.execute()


@pytest.fixture(scope="session")
def sorted_df(df):
    return df.sort_values("id").reset_index(drop=True)


@pytest.fixture(scope="session")
def batting_df(batting):
    return batting.execute(limit=None)


@pytest.fixture(scope="session")
def awards_players_df(awards_players):
    return awards_players.execute(limit=None)


@pytest.fixture(scope="session")
def geo_df(geo):
    if geo is not None:
        return geo.execute(limit=None)
    return None


@pytest.fixture
def temp_table(con) -> str:
    """Return a temporary table name.

    Parameters
    ----------
    con : ibis.backends.base.Client

    Yields
    ------
    name : string
        Random table name for a temporary usage.
    """
    name = util.gen_name("temp_table")
    yield name
    with contextlib.suppress(NotImplementedError):
        con.drop_table(name, force=True)


@pytest.fixture
def temp_table2(con) -> str:
    name = util.gen_name("temp_table2")
    yield name
    with contextlib.suppress(NotImplementedError):
        con.drop_table(name, force=True)


@pytest.fixture
def temp_table_orig(con, temp_table):
    name = f"{temp_table}_orig"
    yield name
    with contextlib.suppress(NotImplementedError):
        con.drop_table(name, force=True)


@pytest.fixture
def temp_view(ddl_con) -> str:
    """Return a temporary view name.

    Parameters
    ----------
    ddl_con : backend connection

    Yields
    ------
    name : string
        Random view name for a temporary usage.
    """
    name = util.gen_name("view")
    yield name
    with contextlib.suppress(NotImplementedError):
        ddl_con.drop_view(name, force=True)


@pytest.fixture
def alternate_current_database(ddl_con, ddl_backend) -> str:
    """Create a temporary database and yield its name. Drops the created
    database upon completion.

    Parameters
    ----------
    ddl_con : ibis.backends.base.Client

    Yields
    ------
    str
    """
    name = util.gen_name("database")
    try:
        ddl_con.create_database(name)
    except AttributeError:
        pytest.skip(f"{ddl_backend.name()} doesn't have a `create_database` method.")
    yield name

    with contextlib.suppress(com.UnsupportedOperationError):
        ddl_con.drop_database(name, force=True)


@pytest.fixture
def test_employee_schema() -> ibis.schema:
    sch = ibis.schema(
        [
            ("first_name", "string"),
            ("last_name", "string"),
            ("department_name", "string"),
            ("salary", "float64"),
        ]
    )

    return sch


@pytest.fixture
def test_employee_data_1():
    df = pd.DataFrame(
        {
            "first_name": ["A", "B", "C"],
            "last_name": ["D", "E", "F"],
            "department_name": ["AA", "BB", "CC"],
            "salary": [100.0, 200.0, 300.0],
        }
    )

    return df


@pytest.fixture
def test_employee_data_2():
    df2 = pd.DataFrame(
        {
            "first_name": ["X", "Y", "Z"],
            "last_name": ["A", "B", "C"],
            "department_name": ["XX", "YY", "ZZ"],
            "salary": [400.0, 500.0, 600.0],
        }
    )

    return df2



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import ibis
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.tests.base import BackendTest
from ibis.conftest import SANDBOXED, WINDOWS

if TYPE_CHECKING:
    from collections.abc import Iterator

    from ibis.backends.base import BaseBackend

TEST_TABLES_GEO = {
    "zones": ibis.schema(
        {
            "zone": "string",
            "LocationID": "int32",
            "borough": "string",
            "geom": "geometry",
            "x_cent": "float32",
            "y_cent": "float32",
        }
    ),
    "lines": ibis.schema(
        {
            "loc_id": "int32",
            "geom": "geometry",
        }
    ),
}


class TestConf(BackendTest):
    supports_map = True
    deps = ("duckdb",)
    stateful = False
    supports_tpch = True
    driver_supports_multiple_statements = True

    def preload(self):
        if not SANDBOXED:
            self.connection._load_extensions(
                ["httpfs", "postgres_scanner", "sqlite_scanner", "spatial"]
            )

    @property
    def ddl_script(self) -> Iterator[str]:
        parquet_dir = self.data_dir / "parquet"
        geojson_dir = self.data_dir / "geojson"
        for table in TEST_TABLES:
            yield (
                f"""
                CREATE OR REPLACE TABLE {table} AS
                SELECT * FROM read_parquet('{parquet_dir / f'{table}.parquet'}')
                """
            )
        if not SANDBOXED:
            for table in TEST_TABLES_GEO:
                yield (
                    f"""
                    CREATE OR REPLACE TABLE {table} AS
                    SELECT * FROM st_read('{geojson_dir / f'{table}.geojson'}')
                    """
                )
            yield (
                """
                CREATE or REPLACE TABLE geo (name VARCHAR, geom GEOMETRY);
                INSERT INTO geo VALUES
                    ('Point', ST_GeomFromText('POINT(-100 40)')),
                    ('Linestring', ST_GeomFromText('LINESTRING(0 0, 1 1, 2 1, 2 2)')),
                    ('Polygon', ST_GeomFromText('POLYGON((0 0, 1 0, 1 1, 0 1, 0 0))'));
                """
            )
        yield from super().ddl_script

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw) -> BaseBackend:
        # use an extension directory per test worker to prevent simultaneous
        # downloads on windows
        #
        # avoid enabling on linux because this adds a lot of time to parallel
        # test runs due to each worker getting its own extensions directory
        if WINDOWS:
            extension_directory = tmpdir.getbasetemp().joinpath("duckdb_extensions")
            extension_directory.mkdir(exist_ok=True)
            kw["extension_directory"] = extension_directory
        return ibis.duckdb.connect(**kw)

    def load_tpch(self) -> None:
        """Load the TPC-H dataset."""
        with self.connection._safe_raw_sql("CALL dbgen(sf=0.17)"):
            pass


@pytest.fixture(scope="session")
def con(data_dir, tmp_path_factory, worker_id):
    return TestConf.load_data(data_dir, tmp_path_factory, worker_id).connection


@pytest.fixture(scope="session")
def gpd():
    pytest.importorskip("shapely")
    return pytest.importorskip("geopandas")


@pytest.fixture(scope="session")
def zones(con, data_dir, gpd):
    return con.read_geo(data_dir / "geojson" / "zones.geojson")


@pytest.fixture(scope="session")
def lines(con, data_dir, gpd):
    return con.read_geo(data_dir / "geojson" / "lines.geojson")


@pytest.fixture(scope="session")
def zones_gdf(data_dir, gpd):
    return gpd.read_file(data_dir / "geojson" / "zones.geojson")


@pytest.fixture(scope="session")
def lines_gdf(data_dir, gpd):
    return gpd.read_file(data_dir / "geojson" / "lines.geojson")


@pytest.fixture(scope="session")
def geotable(con, gpd):
    return con.table("geo")


@pytest.fixture(scope="session")
def gdf(geotable):
    return geotable.execute()



# Copyright 2015 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import operator
from operator import methodcaller

import pytest
from pytest import param

import ibis
from ibis.backends.tests.sql.conftest import to_sql
from ibis.tests.util import assert_decompile_roundtrip

pytestmark = pytest.mark.duckdb


@pytest.fixture(scope="module")
def star1():
    return ibis.table(
        [
            ("c", "int32"),
            ("f", "double"),
            ("foo_id", "string"),
            ("bar_id", "string"),
        ],
        name="star1",
    )


@pytest.fixture(scope="module")
def functional_alltypes():
    return ibis.table(
        {
            "id": "int32",
            "bool_col": "boolean",
            "tinyint_col": "int8",
            "smallint_col": "int16",
            "int_col": "int32",
            "bigint_col": "int64",
            "float_col": "float32",
            "double_col": "float64",
            "date_string_col": "string",
            "string_col": "string",
            "timestamp_col": "timestamp",
            "year": "int32",
            "month": "int32",
        },
        name="functional_alltypes",
    )


@pytest.fixture(scope="module")
def alltypes():
    return ibis.table(
        [
            ("a", "int8"),
            ("b", "int16"),
            ("c", "int32"),
            ("d", "int64"),
            ("e", "float32"),
            ("f", "float64"),
            ("g", "string"),
            ("h", "boolean"),
            ("i", "timestamp"),
            ("j", "date"),
            ("k", "time"),
        ],
        name="alltypes",
    )


@pytest.mark.parametrize("opname", ["ge", "gt", "lt", "le", "eq", "ne"])
def test_comparisons(functional_alltypes, opname, snapshot):
    op = getattr(operator, opname)
    expr = op(functional_alltypes.double_col, 5).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    "expr_fn",
    [
        param(lambda d: (d > 0) & (d < 5), id="and"),
        param(lambda d: (d < 0) | (d > 5), id="or"),
    ],
)
def test_boolean_conjunction(functional_alltypes, expr_fn, snapshot):
    expr = expr_fn(functional_alltypes.double_col).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_between(functional_alltypes, snapshot):
    expr = functional_alltypes.double_col.between(5, 10).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize("method_name", ["isnull", "notnull"])
def test_isnull_notnull(functional_alltypes, method_name, snapshot):
    expr = getattr(functional_alltypes.double_col, method_name)().name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_negate(functional_alltypes, snapshot):
    expr = -(functional_alltypes.double_col > 0)
    snapshot.assert_match(to_sql(expr.name("tmp")), "out.sql")


def test_coalesce(functional_alltypes, snapshot):
    d = functional_alltypes.double_col
    f = functional_alltypes.float_col

    expr = ibis.coalesce((d > 30).ifelse(d, ibis.NA), ibis.NA, f).name("tmp")
    snapshot.assert_match(to_sql(expr.name("tmp")), "out.sql")


def test_named_expr(functional_alltypes, snapshot):
    expr = functional_alltypes[(functional_alltypes.double_col * 2).name("foo")]
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    "expr_fn",
    [
        (lambda r, n: r.inner_join(n, r.r_regionkey == n.n_regionkey)),
        (lambda r, n: r.left_join(n, r.r_regionkey == n.n_regionkey)),
        (lambda r, n: r.outer_join(n, r.r_regionkey == n.n_regionkey)),
        (lambda r, n: r.inner_join(n, r.r_regionkey == n.n_regionkey).select(n)),
        (lambda r, n: r.left_join(n, r.r_regionkey == n.n_regionkey).select(n)),
        (lambda r, n: r.outer_join(n, r.r_regionkey == n.n_regionkey).select(n)),
    ],
    ids=["inner", "left", "outer", "inner_select", "left_select", "outer_select"],
)
def test_joins(region, nation, expr_fn, snapshot):
    expr = expr_fn(region, nation)
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_join_just_materialized(nation, region, customer, snapshot):
    t1 = nation
    t2 = region
    t3 = customer
    joined = t1.inner_join(t2, t1.n_regionkey == t2.r_regionkey).inner_join(
        t3, t1.n_nationkey == t3.c_nationkey
    )  # GH #491

    snapshot.assert_match(to_sql(joined), "out.sql")


def test_full_outer_join(region, nation):
    """Testing full outer join separately due to previous issue with outer join
    resulting in left outer join (issue #1773)"""
    predicate = region.r_regionkey == nation.n_regionkey
    joined = region.outer_join(nation, predicate)
    joined_sql_str = to_sql(joined)
    assert "full" in joined_sql_str.lower()
    assert "left" not in joined_sql_str.lower()


def test_simple_case(simple_case, snapshot):
    expr = simple_case.name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_searched_case(search_case, snapshot):
    expr = search_case.name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_where_simple_comparisons(star1, snapshot):
    t1 = star1
    expr = t1.filter([t1.f > 0, t1.c < t1.f * 2])
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot)


@pytest.mark.parametrize(
    "expr_fn",
    [
        lambda t: t.agg([t["f"].sum().name("total")], [t["foo_id"]]),
        lambda t: t.agg([t["f"].sum().name("total")], ["foo_id", "bar_id"]),
        lambda t: t.agg(
            [t.f.sum().name("total")], by=["foo_id"], having=[t.f.sum() > 10]
        ),
        lambda t: t.agg(
            [t.f.sum().name("total")], by=["foo_id"], having=[t.count() > 100]
        ),
    ],
    ids=["single", "two", "having_sum", "having_count"],
)
def test_aggregate(star1, expr_fn, snapshot):
    expr = expr_fn(star1)
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    "key",
    ["f", ibis.random()],
    ids=["column", "random"],
)
def test_order_by(star1, key, snapshot):
    expr = star1.order_by(key)
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    "expr_fn", [methodcaller("limit", 10), methodcaller("limit", 10, offset=5)]
)
def test_limit(star1, expr_fn, snapshot):
    expr = expr_fn(star1)
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_limit_filter(star1, snapshot):
    expr = star1[star1.f > 0].limit(10)
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_limit_subquery(star1, snapshot):
    expr = star1.limit(10)[lambda x: x.f > 0]
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_cte_factor_distinct_but_equal(alltypes, snapshot):
    t = alltypes
    tt = alltypes.view()

    expr1 = t.group_by("g").aggregate(t.f.sum().name("metric"))
    expr2 = tt.group_by("g").aggregate(tt.f.sum().name("metric")).view()

    expr = expr1.join(expr2, expr1.g == expr2.g)[[expr1]]

    snapshot.assert_match(to_sql(expr), "out.sql")


def test_self_reference_join(star1, snapshot):
    t1 = star1
    t2 = t1.view()
    expr = t1.inner_join(t2, [t1.foo_id == t2.bar_id])[[t1]]

    snapshot.assert_match(to_sql(expr), "out.sql")


def test_self_reference_in_not_exists(functional_alltypes, snapshot):
    t = functional_alltypes
    t2 = t.view()

    cond = (t.string_col == t2.string_col).any()

    semi = t[cond]
    anti = t[-cond]

    snapshot.assert_match(to_sql(semi), "semi.sql")
    snapshot.assert_match(to_sql(anti), "anti.sql")


def test_where_uncorrelated_subquery(foo, bar, snapshot):
    expr = foo[foo.job.isin(bar.job)]

    snapshot.assert_match(to_sql(expr), "out.sql")


def test_where_correlated_subquery(foo, snapshot):
    t1 = foo
    t2 = t1.view()

    stat = t2[t1.dept_id == t2.dept_id].y.mean()
    expr = t1[t1.y > stat]

    snapshot.assert_match(to_sql(expr), "out.sql")


def test_subquery_aliased(star1, star2, snapshot):
    t1 = star1
    t2 = star2

    agged = t1.aggregate([t1.f.sum().name("total")], by=["foo_id"])
    expr = agged.inner_join(t2, [agged.foo_id == t2.foo_id])[agged, t2.value1]

    snapshot.assert_match(to_sql(expr), "out.sql")


def test_lower_projection_sort_key(star1, star2, snapshot):
    t1 = star1
    t2 = star2

    agged = t1.aggregate([t1.f.sum().name("total")], by=["foo_id"])
    expr = agged.inner_join(t2, [agged.foo_id == t2.foo_id])[agged, t2.value1]

    expr2 = expr[expr.total > 100].order_by(ibis.desc("total"))
    snapshot.assert_match(to_sql(expr2), "out.sql")
    assert_decompile_roundtrip(expr2, snapshot)


def test_exists(foo_t, bar_t, snapshot):
    t1 = foo_t
    t2 = bar_t
    cond = (t1.key1 == t2.key1).any()
    e1 = t1[cond]

    snapshot.assert_match(to_sql(e1), "e1.sql")

    cond2 = ((t1.key1 == t2.key1) & (t2.key2 == "foo")).any()
    e2 = t1[cond2]

    snapshot.assert_match(to_sql(e2), "e2.sql")


def test_not_exists(not_exists, snapshot):
    snapshot.assert_match(to_sql(not_exists), "out.sql")


@pytest.mark.parametrize(
    "expr_fn",
    [
        param(lambda t: t.distinct(), id="table_distinct"),
        param(
            lambda t: t["string_col", "int_col"].distinct(), id="projection_distinct"
        ),
        param(
            lambda t: t[t.string_col].distinct(), id="single_column_projection_distinct"
        ),
        param(lambda t: t.int_col.nunique().name("nunique"), id="count_distinct"),
        param(
            lambda t: t.group_by("string_col").aggregate(
                t.int_col.nunique().name("nunique")
            ),
            id="group_by_count_distinct",
        ),
    ],
)
def test_distinct(functional_alltypes, expr_fn, snapshot):
    expr = expr_fn(functional_alltypes)
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_sort_aggregation_translation_failure(functional_alltypes, snapshot):
    # This works around a nuance with our choice to hackishly fuse SortBy
    # after Aggregate to produce a single select statement rather than an
    # inline view.
    t = functional_alltypes

    agg = t.group_by("string_col").aggregate(t.double_col.max().name("foo"))
    expr = agg.order_by(ibis.desc("foo"))

    snapshot.assert_match(to_sql(expr), "out.sql")


def test_where_correlated_subquery_with_join(snapshot):
    # GH3163
    # ibis code
    part = ibis.table([("p_partkey", "int64")], name="part")
    partsupp = ibis.table(
        [
            ("ps_partkey", "int64"),
            ("ps_supplycost", "float64"),
            ("ps_suppkey", "int64"),
        ],
        name="partsupp",
    )
    supplier = ibis.table([("s_suppkey", "int64")], name="supplier")

    q = part.join(partsupp, part.p_partkey == partsupp.ps_partkey)
    q = q[
        part.p_partkey,
        partsupp.ps_supplycost,
    ]
    subq = partsupp.join(supplier, supplier.s_suppkey == partsupp.ps_suppkey)
    subq = subq.select(partsupp.ps_partkey, partsupp.ps_supplycost)
    subq = subq[subq.ps_partkey == q.p_partkey]

    expr = q[q.ps_supplycost == subq.ps_supplycost.min()]

    snapshot.assert_match(to_sql(expr), "out.sql")


def test_mutate_filter_join_no_cross_join(snapshot):
    person = ibis.table(
        [("person_id", "int64"), ("birth_datetime", "timestamp")],
        name="person",
    )
    mutated = person.mutate(age=400)
    expr = mutated.filter(mutated.age <= 40)[mutated.person_id]

    snapshot.assert_match(to_sql(expr), "out.sql")


def test_filter_group_by_agg_with_same_name(snapshot):
    # GH 2907
    t = ibis.table([("int_col", "int32"), ("bigint_col", "int64")], name="t")
    expr = (
        t.group_by("int_col")
        .aggregate(bigint_col=lambda t: t.bigint_col.sum())
        .filter(lambda t: t.bigint_col == 60)
    )
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.fixture
def person():
    return ibis.table(
        dict(id="string", personal="string", family="string"), name="person"
    )


@pytest.fixture
def visited():
    return ibis.table(dict(id="int32", site="string", dated="string"), name="visited")


@pytest.fixture
def survey():
    return ibis.table(
        dict(taken="int32", person="string", quant="string", reading="float32"),
        name="survey",
    )


def test_no_cross_join(person, visited, survey, snapshot):
    expr = person.join(survey, person.id == survey.person).join(
        visited, visited.id == survey.taken
    )
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.fixture
def test1():
    return ibis.table(name="test1", schema=dict(id1="int32", val1="float32"))


@pytest.fixture
def test2():
    return ibis.table(
        name="test2",
        schema=ibis.schema(dict(id2a="int32", id2b="int64", val2="float64")),
    )


@pytest.fixture
def test3():
    return ibis.table(
        name="test3",
        schema=dict(id3="string", val2="float64", dt="timestamp"),
    )


def test_gh_1045(test1, test2, test3, snapshot):
    t1 = test1
    t2 = test2
    t3 = test3

    t3 = t3.mutate(id3=t3.id3.cast("int64"))

    t3 = t3.mutate(t3_val2=t3.id3)
    t4 = t3.join(t2, t2.id2b == t3.id3)

    t1 = t1[[t1[c].name(f"t1_{c}") for c in t1.columns]]

    expr = t1.left_join(t4, t1.t1_id1 == t4.id2a)

    snapshot.assert_match(to_sql(expr), "out.sql")


def test_multi_join(snapshot):
    t1 = ibis.table(dict(x1="int64", y1="int64"), name="t1")
    t2 = ibis.table(dict(x2="int64"), name="t2")
    t3 = ibis.table(dict(x3="int64", y2="int64"), name="t3")
    t4 = ibis.table(dict(x4="int64"), name="t4")

    j1 = t1.join(t2, t1.x1 == t2.x2)
    j2 = t3.join(t4, t3.x3 == t4.x4)
    expr = j1.join(j2, j1.y1 == j2.y2)

    snapshot.assert_match(to_sql(expr), "out.sql")


def test_no_cart_join(snapshot):
    facts = ibis.table(dict(product_id="!int32"), name="facts")
    products = ibis.table(
        dict(
            ancestor_level_name="string",
            ancestor_level_number="int32",
            ancestor_node_sort_order="int64",
            descendant_node_natural_key="int32",
        ),
        name="products",
    )

    products = products.mutate(
        product_level_name=lambda t: ibis.literal("-")
        .lpad(((t.ancestor_level_number - 1) * 7), "-")
        .concat(t.ancestor_level_name)
    )

    predicate = facts.product_id == products.descendant_node_natural_key
    joined = facts.join(products, predicate)

    gb = joined.group_by(products.ancestor_node_sort_order)
    agg = gb.aggregate(n=ibis.literal(1))
    ob = agg.order_by(products.ancestor_node_sort_order)

    snapshot.assert_match(to_sql(ob), "out.sql")


def test_order_by_expr(snapshot):
    t = ibis.table(dict(a="int", b="string"), name="t")
    expr = t[lambda t: t.a == 1].order_by(lambda t: t.b + "a")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_no_cartesian_join(snapshot):
    customers = ibis.table(
        dict(customer_id="int64", first_name="string", last_name="string"),
        name="customers",
    )
    orders = ibis.table(
        dict(order_id="int64", customer_id="int64", order_date="date", status="string"),
        name="orders",
    )
    payments = ibis.table(
        dict(
            payment_id="int64",
            order_id="int64",
            payment_method="string",
            amount="float64",
        ),
        name="payments",
    )

    customer_orders = orders.group_by("customer_id").aggregate(
        first_order=orders.order_date.min(),
        most_recent_order=orders.order_date.max(),
        number_of_orders=orders.order_id.count(),
    )

    customer_payments = (
        payments.left_join(orders, "order_id")
        .group_by(orders.customer_id)
        .aggregate(total_amount=payments.amount.sum())
    )

    final = (
        customers.left_join(customer_orders, "customer_id")
        .drop("customer_id_right")
        .left_join(customer_payments, "customer_id")[
            customers.customer_id,
            customers.first_name,
            customers.last_name,
            customer_orders.first_order,
            customer_orders.most_recent_order,
            customer_orders.number_of_orders,
            customer_payments.total_amount.name("customer_lifetime_value"),
        ]
    )
    snapshot.assert_match(ibis.to_sql(final, dialect="duckdb"), "out.sql")



from __future__ import annotations

from decimal import Decimal

import pandas as pd
import pandas.testing as tm
import pytest
from impala.error import HiveServer2Error
from pytest import param

import ibis
import ibis.expr.types as ir
from ibis import literal as L
from ibis.expr import api


def test_embedded_identifier_quoting(alltypes):
    t = alltypes

    expr = t[[(t.double_col * 2).name("double(fun)")]]["double(fun)"].sum()
    expr.execute()


def test_decimal_metadata(con):
    table = con.table("lineitem")

    expr = table.l_quantity
    assert expr.type().precision == 12
    assert expr.type().scale == 2

    # TODO: what if user impyla version does not have decimal Metadata?


def test_builtins(con, alltypes):
    table = alltypes

    i1 = table.tinyint_col
    i4 = table.int_col
    i8 = table.bigint_col
    d = table.double_col
    s = table.string_col

    exprs = [
        api.now(),
        api.e,
        # hash functions
        i4.hash(),
        d.hash(),
        s.hash(),
        # modulus cases
        i1 % 5,
        i4 % 10,
        20 % i1,
        d % 5,
        i1.fillna(0),
        i4.fillna(0),
        i8.fillna(0),
        i4.to_timestamp("s"),
        i4.to_timestamp("ms"),
        i4.to_timestamp("us"),
        i8.to_timestamp(),
        d.abs(),
        d.cast("decimal(12, 2)"),
        d.cast("int32"),
        d.ceil(),
        d.exp(),
        d.isnull(),
        d.fillna(0),
        d.floor(),
        d.log(),
        d.ln(),
        d.log2(),
        d.log10(),
        d.notnull(),
        d.nullif(0),
        d.round(),
        d.round(2),
        d.round(i1),
        i1.sign(),
        i4.sign(),
        d.sign(),
        # conv
        i1.convert_base(10, 2),
        i4.convert_base(10, 2),
        i8.convert_base(10, 2),
        s.convert_base(10, 2),
        d.sqrt(),
        # nullif cases
        5 / i1.nullif(0),
        5 / i1.nullif(i4),
        5 / i4.nullif(0),
        5 / d.nullif(0),
        api.literal(5).isin([i1, i4, d]),
        # tier and histogram
        d.bucket([0, 10, 25, 50, 100]),
        d.bucket([0, 10, 25, 50], include_over=True),
        d.bucket([0, 10, 25, 50], include_over=True, close_extreme=False),
        d.bucket([10, 25, 50, 100], include_under=True),
        d.histogram(10),
        d.histogram(5, base=10),
        d.histogram(base=10, binwidth=5),
        # coalesce-like cases
        api.coalesce(
            table.int_col, api.null(), table.smallint_col, table.bigint_col, 5
        ),
        api.greatest(table.float_col, table.double_col, 5),
        api.least(table.string_col, "foo"),
        # string stuff
        s.contains("6"),
        s.like("6%"),
        s.re_search(r"[\d]+"),
        s.re_extract(r"[\d]+", 0),
        s.re_replace(r"[\d]+", "a"),
        s.repeat(2),
        s.translate("a", "b"),
        s.find("a"),
        s.lpad(10, "a"),
        s.rpad(10, "a"),
        s.find_in_set(["a"]),
        s.lower(),
        s.upper(),
        s.reverse(),
        s.ascii_str(),
        s.length(),
        s.strip(),
        s.lstrip(),
        s.strip(),
        # strings with int expr inputs
        s.left(i1),
        s.right(i1),
        s.substr(i1, i1 + 2),
        s.repeat(i1),
    ]

    proj_exprs = [expr.name("e%d" % i) for i, expr in enumerate(exprs)]

    projection = table[proj_exprs]
    projection.limit(10).execute()

    _check_impala_output_types_match(con, projection)


def _check_impala_output_types_match(con, table):
    query = ibis.to_sql(table, dialect="impala")
    t = con.sql(query)

    left_schema, right_schema = t.schema(), table.schema()
    for n, left_ty, right_ty in zip(
        left_schema.names, left_schema.types, right_schema.types
    ):
        assert (
            left_ty == right_ty
        ), f"Value for {n} had left type {left_ty} and right type {right_ty}\nquery:\n{query}"


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        # mod cases
        (L(50) % 5, 0),
        (L(50000) % 10, 0),
        (250 % L(50), 0),
        # nullif cases
        (5 / L(50).nullif(0), 0.1),
        (5 / L(50).nullif(L(50000)), 0.1),
        (5 / L(50000).nullif(0), 0.0001),
        (L(50000).fillna(0), 50000),
    ],
)
def test_int_builtins(con, expr, expected):
    result = con.execute(expr)
    assert result == expected, ibis.to_sql(expr, dialect="impala")


@pytest.mark.parametrize(
    ("col", "expected"),
    [
        param("tinyint_col", "int32", id="tinyint"),
        param("smallint_col", "int32", id="smallint"),
        param("int_col", "int32", id="int"),
        param("bigint_col", "int64", id="bigint"),
        param("float_col", "float32", id="float"),
        param("double_col", "float64", id="double"),
        param("timestamp_col", "datetime64[ns]", id="timestamp"),
    ],
)
def test_column_types(alltypes_df, col, expected):
    assert alltypes_df[col].dtype.name == expected


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (L(50000).to_timestamp("s"), pd.to_datetime(50000, unit="s")),
        (L(50000).to_timestamp("ms"), pd.to_datetime(50000, unit="ms")),
        (L(5 * 10**8).to_timestamp(), pd.to_datetime(5 * 10**8, unit="s")),
        (
            ibis.timestamp("2009-05-17 12:34:56").truncate("y"),
            pd.Timestamp("2009-01-01"),
        ),
        (
            ibis.timestamp("2009-05-17 12:34:56").truncate("M"),
            pd.Timestamp("2009-05-01"),
        ),
        (
            ibis.timestamp("2009-05-17 12:34:56").truncate("month"),
            pd.Timestamp("2009-05-01"),
        ),
        (
            ibis.timestamp("2009-05-17 12:34:56").truncate("d"),
            pd.Timestamp("2009-05-17"),
        ),
        (
            ibis.timestamp("2009-05-17 12:34:56").truncate("h"),
            pd.Timestamp("2009-05-17 12:00"),
        ),
        (
            ibis.timestamp("2009-05-17 12:34:56").truncate("m"),
            pd.Timestamp("2009-05-17 12:34"),
        ),
        (
            ibis.timestamp("2009-05-17 12:34:56").truncate("minute"),
            pd.Timestamp("2009-05-17 12:34"),
        ),
    ],
)
def test_timestamp_builtins(con, expr, expected):
    result = con.execute(expr)
    assert result == expected, ibis.to_sql(expr, dialect="impala")


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (L(-5).abs(), 5),
        (L(5.245).cast("int32"), 5),
        (L(5.245).ceil(), 6),
        (L(5.245).isnull(), False),
        (L(5.245).floor(), 5),
        (L(5.245).notnull(), True),
        (L(5.245).round(), 5),
        (L(5.245).round(2), Decimal("5.25")),
        (L(5.245).sign(), 1),
    ],
)
def test_decimal_builtins(con, expr, expected):
    result = con.execute(expr)
    assert result == expected, ibis.to_sql(expr, dialect="impala")


def approx_equal(a, b, eps):
    assert abs(a - b) < eps


@pytest.mark.parametrize(
    ("func", "expected"),
    [
        pytest.param(lambda dc: dc, "5.245", id="id"),
        pytest.param(lambda dc: dc % 5, "0.245", id="mod"),
        pytest.param(lambda dc: dc.fillna(0), "5.245", id="fillna"),
        pytest.param(lambda dc: dc.exp(), "189.6158", id="exp"),
        pytest.param(lambda dc: dc.log(), "1.65728", id="log"),
        pytest.param(lambda dc: dc.log2(), "2.39094", id="log2"),
        pytest.param(lambda dc: dc.log10(), "0.71975", id="log10"),
        pytest.param(lambda dc: dc.sqrt(), "2.29019", id="sqrt"),
        pytest.param(lambda dc: dc.fillna(0), "5.245", id="zero_ifnull"),
        pytest.param(lambda dc: -dc, "-5.245", id="neg"),
    ],
)
def test_decimal_builtins_2(con, func, expected):
    dc = L("5.245").cast("decimal(12, 5)")
    expr = func(dc)
    result = con.execute(expr)
    tol = Decimal("0.0001")
    approx_equal(Decimal(result), Decimal(expected), tol)


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (L("abcd").length(), 4),
        (L("ABCD").lower(), "abcd"),
        (L("abcd").upper(), "ABCD"),
        (L("abcd").reverse(), "dcba"),
        (L("abcd").ascii_str(), 97),
        (L("   a   ").strip(), "a"),
        (L("   a   ").lstrip(), "a   "),
        (L("   a   ").rstrip(), "   a"),
        (L("abcd").substr(0, 2), "ab"),
        (L("abcd").left(2), "ab"),
        (L("abcd").right(2), "cd"),
        (L("abcd").repeat(2), "abcdabcd"),
        # global replace not available in Impala yet
        # (L('aabbaabbaa').replace('bb', 'B'), 'aaBaaBaa'),
        (L("0123").translate("012", "abc"), "abc3"),
        (L("abcd").find("a"), 0),
        (L("baaaab").find("b", 2), 5),
        (L("abcd").lpad(1, "-"), "a"),
        (L("abcd").lpad(5), " abcd"),
        (L("abcd").rpad(1, "-"), "a"),
        (L("abcd").rpad(5), "abcd "),
        (L("abcd").find_in_set(["a", "b", "abcd"]), 2),
        (L(", ").join(["a", "b"]), "a, b"),
        (L("abcd").like("a%"), True),
        (L("abcd").re_search("[a-z]"), True),
        (L("abcd").re_extract("[a-z]", 0), "a"),
        (L("abcd").re_replace("(b)", "2"), "a2cd"),
    ],
)
def test_string_functions(con, expr, expected):
    result = con.execute(expr)
    assert result == expected


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (L("https://www.cloudera.com").host(), "www.cloudera.com"),
        (
            L("https://www.youtube.com/watch?v=kEuEcWfewf8&t=10").query("v"),
            "kEuEcWfewf8",
        ),
    ],
)
def test_parse_url(con, expr, expected):
    result = con.execute(expr)
    assert result == expected


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        (L(7) / 2, 3.5),
        (L(7) // 2, 3),
        (L(7).floordiv(2), 3),
        (L(2).rfloordiv(7), 3),
    ],
)
def test_div_floordiv(con, expr, expected):
    result = con.execute(expr)
    assert result == expected


def test_filter_predicates(con):
    t = con.table("nation")

    predicates = [
        lambda x: x.n_name.lower().like("%ge%"),
        lambda x: x.n_name.lower().contains("ge"),
        lambda x: x.n_name.lower().rlike(".*ge.*"),
    ]

    expr = t
    for pred in predicates:
        expr = expr[pred(expr)].select(expr)

    expr.execute()


def test_histogram_value_counts(alltypes):
    t = alltypes
    expr = t.double_col.histogram(10).value_counts()
    expr.execute()


def test_casted_expr_impala_bug(alltypes):
    # Per GH #396. Prior to Impala 2.3.0, there was a bug in the query
    # planner that caused this expression to fail
    expr = alltypes.string_col.cast("double").value_counts()
    expr.execute()


def test_decimal_timestamp_builtins(con):
    table = con.table("lineitem")

    dc = table.l_quantity
    ts = table.l_receiptdate.cast("timestamp")

    exprs = [
        dc % 10,
        dc + 5,
        dc + dc,
        dc / 2,
        dc * 2,
        dc**2,
        dc.cast("double"),
        api.ifelse(table.l_discount > 0, dc * table.l_discount, api.NA),
        dc.fillna(0),
        ts < (ibis.now() + ibis.interval(months=3)),
        ts < (ibis.timestamp("2005-01-01") + ibis.interval(months=3)),
        # hashing
        dc.hash(),
        ts.hash(),
        # truncate
        ts.truncate("y"),
        ts.truncate("q"),
        ts.truncate("month"),
        ts.truncate("d"),
        ts.truncate("w"),
        ts.truncate("h"),
        ts.truncate("minute"),
    ]

    timestamp_fields = [
        "years",
        "months",
        "days",
        "hours",
        "minutes",
        "seconds",
        "weeks",
    ]
    for field in timestamp_fields:
        if hasattr(ts, field):
            exprs.append(getattr(ts, field)())

        offset = ibis.interval(**{field: 2})
        exprs.append(ts + offset)
        exprs.append(ts - offset)

    proj_exprs = [expr.name("e%d" % i) for i, expr in enumerate(exprs)]

    projection = table[proj_exprs].limit(10)
    projection.execute()


def test_timestamp_scalar_in_filter(alltypes):
    # #310
    table = alltypes

    expr = table.filter(
        [
            table.timestamp_col
            < (ibis.timestamp("2010-01-01") + ibis.interval(months=3)),
            table.timestamp_col < (ibis.now() + ibis.interval(days=10)),
        ]
    ).count()
    expr.execute()


def test_aggregations(alltypes):
    table = alltypes.limit(100)

    d = table.double_col
    s = table.string_col

    cond = table.string_col.isin(["1", "7"])

    exprs = [
        table.bool_col.count(),
        d.sum(),
        d.mean(),
        d.min(),
        d.max(),
        s.approx_nunique(),
        d.approx_median(),
        s.group_concat(),
        d.std(),
        d.std(how="pop"),
        d.var(),
        d.var(how="pop"),
        table.bool_col.any(),
        table.bool_col.notany(),
        -table.bool_col.any(),
        table.bool_col.all(),
        table.bool_col.notall(),
        -table.bool_col.all(),
        table.bool_col.count(where=cond),
        d.sum(where=cond),
        d.mean(where=cond),
        d.min(where=cond),
        d.max(where=cond),
        d.std(where=cond),
        d.var(where=cond),
    ]

    metrics = [expr.name("e%d" % i) for i, expr in enumerate(exprs)]

    agged_table = table.aggregate(metrics)
    agged_table.execute()


def test_analytic_functions(alltypes):
    t = alltypes.limit(1000)

    g = t.group_by("string_col").order_by("double_col")
    f = t.float_col

    exprs = [
        f.lag(),
        f.lead(),
        f.rank(),
        f.dense_rank(),
        f.percent_rank(),
        f.ntile(buckets=7),
        f.first(),
        f.last(),
        f.first().over(ibis.window(preceding=10)),
        f.first().over(ibis.window(following=10)),
        ibis.row_number(),
        f.cumsum(),
        f.cummean(),
        f.cummin(),
        f.cummax(),
        # boolean cumulative reductions
        (f == 0).cumany(),
        (f == 0).cumall(),
        f.sum(),
        f.mean(),
        f.min(),
        f.max(),
    ]

    proj_exprs = [expr.name("e%d" % i) for i, expr in enumerate(exprs)]

    proj_table = g.mutate(proj_exprs)
    proj_table.execute()


def test_anti_join_self_reference_works(con, alltypes):
    t = alltypes.limit(100)
    t2 = t.view()
    case = t[-((t.string_col == t2.string_col).any())]
    con.explain(case)


def test_tpch_self_join_failure(con):
    region = con.table("region")
    nation = con.table("nation")
    customer = con.table("customer")
    orders = con.table("orders")

    fields_of_interest = [
        region.r_name.name("region"),
        nation.n_name.name("nation"),
        orders.o_totalprice.name("amount"),
        orders.o_orderdate.cast("timestamp").name("odate"),
    ]

    joined_all = (
        region.join(nation, region.r_regionkey == nation.n_regionkey)
        .join(customer, customer.c_nationkey == nation.n_nationkey)
        .join(orders, orders.o_custkey == customer.c_custkey)[fields_of_interest]
    )

    year = joined_all.odate.year().name("year")
    total = joined_all.amount.sum().cast("double").name("total")
    annual_amounts = joined_all.group_by(["region", year]).aggregate(total)

    current = annual_amounts
    prior = annual_amounts.view()

    yoy_change = (current.total - prior.total).name("yoy_change")
    yoy = current.join(
        prior,
        ((current.region == prior.region) & (current.year == (prior.year - 1))),
    )[current.region, current.year, yoy_change]

    # no analysis failure
    con.explain(yoy)


def test_tpch_correlated_subquery_failure(con):
    # #183 and other issues
    region = con.table("region")
    nation = con.table("nation")
    customer = con.table("customer")
    orders = con.table("orders")

    fields_of_interest = [
        customer,
        region.r_name.name("region"),
        orders.o_totalprice.name("amount"),
        orders.o_orderdate.cast("timestamp").name("odate"),
    ]

    tpch = (
        region.join(nation, region.r_regionkey == nation.n_regionkey)
        .join(customer, customer.c_nationkey == nation.n_nationkey)
        .join(orders, orders.o_custkey == customer.c_custkey)[fields_of_interest]
    )

    t2 = tpch.view()
    conditional_avg = t2[(t2.region == tpch.region)].amount.mean()
    amount_filter = tpch.amount > conditional_avg

    expr = tpch[amount_filter].limit(0)

    # impala can't plan this because its correlated subquery implementation is
    # broken: it cannot detect the outer reference inside the inner query
    with pytest.raises(HiveServer2Error, match="Could not resolve .+ reference"):
        con.explain(expr)


def test_non_equijoin(con):
    t = con.table("functional_alltypes").limit(100)
    t2 = t.view()

    expr = t.join(t2, t.tinyint_col < t2.timestamp_col.minute()).count()

    # it works
    expr.execute()


def test_char_varchar_types(con):
    sql = """\
SELECT CAST(string_col AS varchar(20)) AS varchar_col,
   CAST(string_col AS CHAR(5)) AS char_col
FROM functional_alltypes"""

    t = con.sql(sql)

    assert isinstance(t.varchar_col, ir.StringColumn)
    assert isinstance(t.char_col, ir.StringColumn)


def test_unions_with_ctes(con, alltypes):
    t = alltypes

    expr1 = t.group_by(["tinyint_col", "string_col"]).aggregate(
        t.double_col.sum().name("metric")
    )
    expr2 = expr1.view()

    join1 = expr1.join(expr2, expr1.string_col == expr2.string_col)[[expr1]]
    join2 = join1.view()

    expr = join1.union(join2)
    con.explain(expr)


@pytest.mark.parametrize(
    ("left", "right", "expected"),
    [
        (ibis.NA.cast("int64"), ibis.NA.cast("int64"), True),
        (L(1), L(1), True),
        (ibis.NA.cast("int64"), L(1), False),
        (L(1), ibis.NA.cast("int64"), False),
        (L(0), L(1), False),
        (L(1), L(0), False),
    ],
)
def test_identical_to(con, left, right, expected):
    expr = left.identical_to(right)
    result = con.execute(expr)
    assert result == expected


def test_not(alltypes):
    t = alltypes.limit(10)
    expr = t.select(double_col=~t.double_col.isnull())
    result = expr.execute().double_col
    expected = ~t.execute().double_col.isnull()
    tm.assert_series_equal(result, expected)


def test_where_with_timestamp(snapshot):
    t = ibis.table(
        [("uuid", "string"), ("ts", "timestamp"), ("search_level", "int64")],
        name="t",
    )
    expr = t.group_by(t.uuid).aggregate(min_date=t.ts.min(where=t.search_level == 1))
    snapshot.assert_match(ibis.impala.compile(expr), "out.sql")


def test_filter_with_analytic(snapshot):
    x = ibis.table(ibis.schema([("col", "int32")]), "x")
    with_filter_col = x[x.columns + [ibis.null().name("filter")]]
    filtered = with_filter_col[with_filter_col["filter"].isnull()]
    subquery = filtered[filtered.columns]

    with_analytic = subquery[["col", subquery.count().name("analytic")]]
    expr = with_analytic[with_analytic.columns]

    snapshot.assert_match(ibis.impala.compile(expr), "out.sql")


def test_named_from_filter_group_by(snapshot):
    t = ibis.table([("key", "string"), ("value", "double")], name="t0")
    gb = t.filter(t.value == 42).group_by(t.key)
    sum_expr = lambda t: (t.value + 1 + 2 + 3).sum()
    expr = gb.aggregate(abc=sum_expr)
    snapshot.assert_match(ibis.impala.compile(expr), "abc.sql")

    expr = gb.aggregate(foo=sum_expr)
    snapshot.assert_match(ibis.impala.compile(expr), "foo.sql")


def test_nunique_where(snapshot):
    t = ibis.table([("key", "string"), ("value", "double")], name="t0")
    expr = t.key.nunique(where=t.value >= 1.0)
    snapshot.assert_match(ibis.impala.compile(expr), "out.sql")



from __future__ import annotations

import decimal

import pandas as pd
import pytest
import pytz
from pytest import param

import ibis
import ibis.expr.datatypes as dt
from ibis.backends.conftest import is_older_than

pytest.importorskip("dask.dataframe")
from dask.dataframe.utils import tm  # noqa: E402

TIMESTAMP = "2022-03-13 06:59:10.467417"


@pytest.mark.parametrize("from_", ["plain_float64", "plain_int64"])
@pytest.mark.parametrize(
    ("to", "expected"),
    [
        ("float16", "float16"),
        ("float32", "float32"),
        ("float64", "float64"),
        ("double", "float64"),
        ("float", "float64"),
        ("int8", "int8"),
        ("int16", "int16"),
        ("int32", "int32"),
        ("int64", "int64"),
        ("string", "object"),
    ],
)
def test_cast_numeric(t, df, from_, to, expected):
    c = t[from_].cast(to)
    result = c.execute()
    assert str(result.dtype) == expected


@pytest.mark.parametrize("from_", ["float64_as_strings", "int64_as_strings"])
@pytest.mark.parametrize(
    ("to", "expected"), [("double", "float64"), ("string", "object")]
)
def test_cast_string(t, df, from_, to, expected):
    c = t[from_].cast(to)
    result = c.execute()
    assert str(result.dtype) == expected


@pytest.mark.parametrize(
    ("to", "expected"),
    [
        pytest.param(
            "string",
            "object",
            marks=pytest.mark.skipif(
                is_older_than("pandas", "2.1.0"), reason="raises a NotImplementedError"
            ),
        ),
        ("int64", "int64"),
        (
            dt.Timestamp("America/Los_Angeles"),
            "datetime64[ns, America/Los_Angeles]",
        ),
        (
            "timestamp('America/Los_Angeles')",
            "datetime64[ns, America/Los_Angeles]",
        ),
    ],
)
@pytest.mark.parametrize(
    "column",
    ["plain_datetimes_naive", "plain_datetimes_ny", "plain_datetimes_utc"],
)
def test_cast_timestamp_column(t, df, column, to, expected):
    c = t[column].cast(to)
    result = c.execute()
    assert str(result.dtype) == expected


@pytest.mark.parametrize(
    ("to", "expected"),
    [
        pytest.param(
            "string",
            str,
            marks=pytest.mark.skipif(
                is_older_than("pandas", "2.1.0"), reason="raises a NotImplementedError"
            ),
        ),
        ("int64", lambda x: pd.Timestamp(x).value // int(1e9)),
        ("double", lambda x: float(pd.Timestamp(x).value // int(1e9))),
        (
            dt.Timestamp("America/Los_Angeles"),
            lambda x: x.tz_localize(tz="America/Los_Angeles"),
        ),
    ],
)
def test_cast_timestamp_scalar_naive(con, to, expected):
    literal_expr = ibis.literal(pd.Timestamp(TIMESTAMP))
    value = literal_expr.cast(to)
    result = con.execute(value)
    raw = con.execute(literal_expr)
    assert result == expected(raw)


@pytest.mark.parametrize(
    ("to", "expected"),
    [
        pytest.param(
            "string",
            str,
            marks=pytest.mark.skipif(
                is_older_than("pandas", "2.1.0"), reason="raises a NotImplementedError"
            ),
        ),
        ("int64", lambda x: pd.Timestamp(x).value // int(1e9)),
        param("double", float, marks=pytest.mark.notimpl(["dask"])),
        (
            dt.Timestamp("America/Los_Angeles"),
            lambda x: x.astimezone(tz=pytz.timezone("America/Los_Angeles")),
        ),
    ],
)
@pytest.mark.parametrize("tz", ["UTC", "America/New_York"])
def test_cast_timestamp_scalar(to, expected, tz, con):
    literal_expr = ibis.literal(pd.Timestamp(TIMESTAMP).tz_localize(tz))
    value = literal_expr.cast(to)
    result = con.execute(value)
    raw = con.execute(literal_expr)
    assert result == expected(raw)


def test_timestamp_with_timezone_is_inferred_correctly(t):
    assert t.plain_datetimes_naive.type().equals(dt.timestamp)
    assert t.plain_datetimes_ny.type().equals(dt.Timestamp("America/New_York"))
    assert t.plain_datetimes_utc.type().equals(dt.Timestamp("UTC"))


@pytest.mark.parametrize(
    "column",
    ["plain_datetimes_naive", "plain_datetimes_ny", "plain_datetimes_utc"],
)
def test_cast_date(t, df, column):
    expr = t[column].cast("date")
    result = expr.execute()
    expected = (
        df[column]
        .dt.normalize()
        .map(lambda x: x.date())
        .compute()
        .rename(expr.get_name())
    )
    tm.assert_series_equal(result, expected, check_index=False)


@pytest.mark.parametrize("type", [dt.Decimal(9, 2), dt.Decimal(12, 3)])
def test_cast_to_decimal(t, pandas_df, type):
    expr = t.float64_as_strings.cast(type)
    result = expr.execute()
    context = decimal.Context(prec=type.precision)
    expected = pandas_df.float64_as_strings.apply(
        lambda x: context.create_decimal(x).quantize(
            decimal.Decimal(
                "{}.{}".format("0" * (type.precision - type.scale), "0" * type.scale)
            )
        )
    )
    tm.assert_series_equal(result, expected, check_names=False)
    assert all(
        abs(element.as_tuple().exponent) == type.scale for element in result.values
    )
    assert all(
        1 <= len(element.as_tuple().digits) <= type.precision
        for element in result.values
    )



from __future__ import annotations

# Copyright 2015 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

import ibis
from ibis.tests.expr.mocks import MockBackend


@pytest.fixture
def schema():
    return [
        ("a", "int8"),
        ("b", "int16"),
        ("c", "int32"),
        ("d", "int64"),
        ("e", "float32"),
        ("f", "float64"),
        ("g", "string"),
        ("h", "boolean"),
        ("i", "timestamp"),
        ("j", "date"),
        ("k", "time"),
    ]


@pytest.fixture
def table(schema):
    return ibis.table(schema, name="table")


@pytest.fixture(params=list("abcdh"))
def int_col(request):
    return request.param


@pytest.fixture(params=list("ef"))
def float_col(request):
    return request.param


@pytest.fixture(params=list("abcdefh"))
def numeric_col(request):
    return request.param


@pytest.fixture(params=list("g"))
def string_col(request):
    return request.param


@pytest.fixture
def con():
    return MockBackend()


@pytest.fixture
def alltypes(con):
    return con.table("alltypes")


@pytest.fixture
def functional_alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture
def lineitem(con):
    return con.table("tpch_lineitem")



from __future__ import annotations

import ibis

from .conftest import tpch_test


@tpch_test
def test_tpc_h13(customer, orders):
    """Customer Distribution Query (Q13)

    This query seeks relationships between customers and the size of their
    orders."""

    WORD1 = "special"
    WORD2 = "requests"

    innerq = customer
    innerq = innerq.left_join(
        orders,
        (customer.c_custkey == orders.o_custkey)
        & ~orders.o_comment.like(f"%{WORD1}%{WORD2}%"),
    )
    innergq = innerq.group_by([innerq.c_custkey])
    innerq = innergq.aggregate(c_count=innerq.o_orderkey.count())

    gq = innerq.group_by([innerq.c_count])
    q = gq.aggregate(custdist=innerq.count())

    q = q.order_by([ibis.desc(q.custdist), ibis.desc(q.c_count)])
    return q



from __future__ import annotations

import pytest

import ibis
import ibis.common.exceptions as exc
import ibis.selectors as s


def test_individual_columns():
    t = ibis.table(dict(x="int", y="int"))
    assert t.relocate("x", after="y").columns == list("yx")
    assert t.relocate("y", before="x").columns == list("yx")


def test_move_blocks():
    t = ibis.table(dict(x="int", a="string", y="int", b="string"))
    assert t.relocate(s.of_type("string")).columns == list("abxy")
    assert t.relocate(s.of_type("string"), after=s.numeric()).columns == list("xyab")


def test_duplicates_not_renamed():
    t = ibis.table(dict(x="int", y="int"))
    assert t.relocate("y", s.numeric()).columns == list("yx")
    assert t.relocate("y", s.numeric(), "y").columns == list("yx")


def test_keep_non_contiguous_variables():
    t = ibis.table(dict.fromkeys("abcde", "int"))
    assert t.relocate("b", after=s.c("a", "c", "e")).columns == list("acdeb")
    assert t.relocate("e", before=s.c("b", "d")).columns == list("aebcd")


def test_before_after_does_not_move_to_front():
    t = ibis.table(dict(x="int", y="int"))
    assert t.relocate("y").columns == list("yx")


def test_only_one_of_before_and_after():
    t = ibis.table(dict(x="int", y="int", z="int"))

    with pytest.raises(exc.IbisInputError, match="Cannot specify both"):
        t.relocate("z", before="x", after="y")


def test_respects_order():
    t = ibis.table(dict.fromkeys("axbzy", "int"))
    assert t.relocate("x", "y", "z", before="x").columns == list("axyzb")
    assert t.relocate("x", "y", "z", before=s.last()).columns == list("abxyz")
    assert t.relocate("x", "a", "z").columns == list("xazby")


def test_relocate_can_rename():
    t = ibis.table(dict(a="int", b="int", c="int", d="string", e="string", f=r"string"))
    assert t.relocate(ffff="f").columns == ["ffff", *"abcde"]
    assert t.relocate(ffff="f", before="c").columns == [*"ab", "ffff", *"cde"]
    assert t.relocate(ffff="f", after="c").columns == [*"abc", "ffff", *"de"]


def test_retains_last_duplicate_when_renaming_and_moving():
    t = ibis.table(dict(x="int"))
    assert t.relocate(a="x", b="x").columns == ["b"]

    # TODO: test against .rename once that's implemented

    t = ibis.table(dict(x="int", y="int"))
    assert t.relocate(a="x", b="y", c="x").columns == list("bc")


def test_everything():
    t = ibis.table(dict(w="int", x="int", y="int", z="int"))
    assert t.relocate("y", "z", before=s.all()).columns == list("yzwx")
    assert t.relocate("y", "z", after=s.all()).columns == list("wxyz")


def test_moves_to_front_with_no_before_and_no_after():
    t = ibis.table(dict(x="int", y="int", z="int"))
    assert t.relocate("z", "y").columns == list("zyx")


def test_empty_before_moves_to_front():
    t = ibis.table(dict(x="int", y="int", z="int"))
    assert t.relocate("y", before=s.of_type("string")).columns == list("yxz")


def test_empty_after_moves_to_end():
    t = ibis.table(dict(x="int", y="int", z="int"))
    assert t.relocate("y", after=s.of_type("string")).columns == list("xzy")


def test_no_arguments():
    t = ibis.table(dict(x="int", y="int", z="int"))
    with pytest.raises(exc.IbisInputError, match="At least one selector"):
        assert t.relocate()



from __future__ import annotations

import contextlib
import functools
import sqlite3
from typing import TYPE_CHECKING, Any

import sqlglot as sg
import sqlglot.expressions as sge

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends.base import UrlFromPath
from ibis.backends.base.sqlglot import SQLGlotBackend
from ibis.backends.base.sqlglot.compiler import C, F
from ibis.backends.sqlite.compiler import SQLiteCompiler
from ibis.backends.sqlite.converter import SQLitePandasData
from ibis.backends.sqlite.udf import ignore_nulls, register_all

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping
    from pathlib import Path

    import pandas as pd
    import pyarrow as pa


@functools.cache
def _init_sqlite3():
    import pandas as pd

    # required to support pandas Timestamp's from user input
    sqlite3.register_adapter(pd.Timestamp, pd.Timestamp.isoformat)


def _quote(name: str) -> str:
    return sg.to_identifier(name, quoted=True).sql("sqlite")


class Backend(SQLGlotBackend, UrlFromPath):
    name = "sqlite"
    compiler = SQLiteCompiler()
    supports_python_udfs = True

    @property
    def current_database(self) -> str:
        return "main"

    @property
    def version(self) -> str:
        return sqlite3.sqlite_version

    def do_connect(
        self,
        database: str | Path | None = None,
        type_map: dict[str, str | dt.DataType] | None = None,
    ) -> None:
        """Create an Ibis client connected to a SQLite database.

        Multiple database files can be accessed using the `attach()` method.

        Parameters
        ----------
        database
            File path to the SQLite database file. If `None`, creates an
            in-memory transient database and you can use attach() to add more
            files
        type_map
            An optional mapping from a string name of a SQLite "type" to the
            corresponding ibis DataType that it represents. This can be used
            to override schema inference for a given SQLite database.

        Examples
        --------
        >>> import ibis
        >>> ibis.sqlite.connect("path/to/my/sqlite.db")

        """
        _init_sqlite3()

        if type_map:
            self._type_map = {k.lower(): ibis.dtype(v) for k, v in type_map.items()}
        else:
            self._type_map = {}

        self.con = sqlite3.connect(":memory:" if database is None else database)

        register_all(self.con)
        self.con.execute("PRAGMA case_sensitive_like=ON")

    def raw_sql(self, query: str | sg.Expression, **kwargs: Any) -> Any:
        if not isinstance(query, str):
            query = query.sql(dialect=self.name)
        return self.con.execute(query, **kwargs)

    @contextlib.contextmanager
    def _safe_raw_sql(self, *args, **kwargs):
        with contextlib.closing(self.raw_sql(*args, **kwargs)) as result:
            yield result

    @contextlib.contextmanager
    def begin(self):
        cur = self.con.cursor()
        try:
            yield cur
        except Exception:
            self.con.rollback()
            raise
        else:
            self.con.commit()
        finally:
            cur.close()

    def list_databases(self, like: str | None = None) -> list[str]:
        with self._safe_raw_sql("SELECT name FROM pragma_database_list()") as cur:
            results = [r[0] for r in cur.fetchall()]

        return sorted(self._filter_with_like(results, like))

    def list_tables(
        self,
        like: str | None = None,
        database: str | None = None,
    ) -> list[str]:
        if database is None:
            database = "main"

        sql = (
            sg.select("name")
            .from_(F.pragma_table_list())
            .where(
                C.schema.eq(database),
                C.type.isin("table", "view"),
                ~(
                    C.name.isin(
                        "sqlite_schema",
                        "sqlite_master",
                        "sqlite_temp_schema",
                        "sqlite_temp_master",
                    )
                ),
            )
            .sql(self.name)
        )
        with self._safe_raw_sql(sql) as cur:
            results = [r[0] for r in cur.fetchall()]

        return sorted(self._filter_with_like(results, like))

    def _parse_type(self, typ: str, nullable: bool) -> dt.DataType:
        typ = typ.lower()
        try:
            out = self._type_map[typ]
        except KeyError:
            return self.compiler.type_mapper.from_string(typ, nullable=nullable)
        else:
            return out.copy(nullable=nullable)

    def _inspect_schema(
        self, cur: sqlite3.Cursor, table_name: str, database: str | None = None
    ) -> Iterator[tuple[str, dt.DataType]]:
        if database is None:
            database = "main"

        quoted_db = _quote(database)
        quoted_table = _quote(table_name)

        sql = f'SELECT name, type, "notnull" FROM {quoted_db}.pragma_table_info({quoted_table})'
        cur.execute(sql)
        rows = cur.fetchall()
        if not rows:
            raise com.IbisError(f"Table not found: {table_name!r}")

        table_info = {name: (typ, not notnull) for name, typ, notnull in rows}

        # if no type info was returned for a column, fetch the type of the
        # first row and assume that matches the rest of the rows
        unknown = [name for name, (typ, _) in table_info.items() if not typ]
        if unknown:
            queries = ", ".join(f"typeof({_quote(name)})" for name in unknown)
            cur.execute(f"SELECT {queries} FROM {quoted_db}.{quoted_table} LIMIT 1")
            row = cur.fetchone()
            if row is not None:
                for name, typ in zip(unknown, row):
                    _, nullable = table_info[name]
                    table_info[name] = (typ, nullable)
            else:
                raise com.IbisError(f"Failed to infer types for columns {unknown}")

        for name, (typ, nullable) in table_info.items():
            yield name, self._parse_type(typ, nullable)

    def get_schema(
        self, table_name: str, schema: str | None = None, database: str | None = None
    ) -> sch.Schema:
        """Compute the schema of a `table`.

        Parameters
        ----------
        table_name
            May **not** be fully qualified. Use `database` if you want to
            qualify the identifier.
        schema
            Schema name. Unused for sqlite.
        database
            Database name

        Returns
        -------
        sch.Schema
            Ibis schema

        """
        if schema is not None:
            raise TypeError("sqlite doesn't support `schema`, use `database` instead")
        with self.begin() as cur:
            return sch.Schema.from_tuples(
                self._inspect_schema(cur, table_name, database)
            )

    def _metadata(self, query: str) -> Iterator[tuple[str, dt.DataType]]:
        with self.begin() as cur:
            # create a view that should only be visible in this transaction
            view = util.gen_name("ibis_sqlite_metadata")
            cur.execute(f"CREATE TEMPORARY VIEW {view} AS {query}")

            yield from self._inspect_schema(cur, view, database="temp")

            # drop the view when we're done with it
            cur.execute(f"DROP VIEW IF EXISTS {view}")

    def _fetch_from_cursor(
        self, cursor: sqlite3.Cursor, schema: sch.Schema
    ) -> pd.DataFrame:
        import pandas as pd

        df = pd.DataFrame.from_records(cursor, columns=schema.names, coerce_float=True)
        return SQLitePandasData.convert_table(df, schema)

    @util.experimental
    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1_000_000,
        **_: Any,
    ) -> pa.ipc.RecordBatchReader:
        import pyarrow as pa

        self._run_pre_execute_hooks(expr)

        schema = expr.as_table().schema()
        with self._safe_raw_sql(
            self.compile(expr, limit=limit, params=params)
        ) as cursor:
            df = self._fetch_from_cursor(cursor, schema)
        table = pa.Table.from_pandas(
            df, schema=schema.to_pyarrow(), preserve_index=False
        )
        return table.to_reader(max_chunksize=chunk_size)

    def _generate_create_table(self, table: sge.Table, schema: sch.Schema):
        column_defs = [
            sge.ColumnDef(
                this=sg.to_identifier(colname, quoted=self.compiler.quoted),
                kind=self.compiler.type_mapper.from_ibis(typ),
                constraints=(
                    None
                    if typ.nullable
                    else [sge.ColumnConstraint(kind=sge.NotNullColumnConstraint())]
                ),
            )
            for colname, typ in schema.items()
        ]

        target = sge.Schema(this=table, expressions=column_defs)

        return sge.Create(kind="TABLE", this=target)

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        # only register if we haven't already done so
        if op.name not in self.list_tables(database="temp"):
            table = sg.table(op.name, quoted=self.compiler.quoted, catalog="temp")
            create_stmt = self._generate_create_table(table, op.schema).sql(self.name)
            df = op.data.to_frame()

            data = df.itertuples(index=False)
            cols = ", ".join(_quote(col) for col in op.schema.keys())
            specs = ", ".join(["?"] * len(op.schema))
            insert_stmt = (
                f"INSERT INTO {table.sql(self.name)} ({cols}) VALUES ({specs})"
            )

            with self.begin() as cur:
                cur.execute(create_stmt)
                cur.executemany(insert_stmt, data)

    def _define_udf_translation_rules(self, expr):
        """No-op, these are defined in the compiler."""

    def _register_udfs(self, expr: ir.Expr) -> None:
        import ibis.expr.operations as ops

        con = self.con

        for udf_node in expr.op().find(ops.ScalarUDF):
            compile_func = getattr(
                self, f"_compile_{udf_node.__input_type__.name.lower()}_udf"
            )
            registration_func = compile_func(udf_node)
            if registration_func is not None:
                registration_func(con)

    def _compile_python_udf(self, udf_node: ops.ScalarUDF) -> None:
        name = type(udf_node).__name__
        nargs = len(udf_node.__signature__.parameters)
        func = udf_node.__func__

        def check_dtype(dtype, name=None):
            if not (
                dtype.is_string()
                or dtype.is_binary()
                or dtype.is_numeric()
                or dtype.is_boolean()
            ):
                label = "return value" if name is None else f"argument `{name}`"
                raise com.IbisTypeError(
                    "SQLite only supports strings, bytes, booleans and numbers as UDF input and output, "
                    f"{label} has unsupported type {dtype}"
                )

        for argname, arg in zip(udf_node.argnames, udf_node.args):
            check_dtype(arg.dtype, argname)
        check_dtype(udf_node.dtype)

        def register_udf(con):
            return con.create_function(name, nargs, ignore_nulls(func))

        return register_udf

    def attach(self, name: str, path: str | Path) -> None:
        """Connect another SQLite database file to the current connection.

        Parameters
        ----------
        name
            Database name within SQLite
        path
            Path to sqlite3 database files

        Examples
        --------
        >>> con1 = ibis.sqlite.connect("original.db")
        >>> con2 = ibis.sqlite.connect("new.db")
        >>> con1.attach("new", "new.db")
        >>> con1.list_tables(database="new")

        """
        with self.begin() as cur:
            cur.execute(f"ATTACH DATABASE {str(path)!r} AS {_quote(name)}")

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: ibis.Schema | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ):
        """Create a table in SQLite.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists

        """
        if schema is None and obj is None:
            raise ValueError("Either `obj` or `schema` must be specified")

        if schema is not None:
            schema = ibis.schema(schema)

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                obj = ibis.memtable(obj)

            self._run_pre_execute_hooks(obj)

            insert_query = self._to_sqlglot(obj)
        else:
            insert_query = None

        if temp:
            if database not in (None, "temp"):
                raise ValueError(
                    "SQLite doesn't support creating temporary tables in an explicit database"
                )
            else:
                database = "temp"

        if overwrite:
            created_table = sg.table(
                util.gen_name(f"{self.name}_table"),
                catalog=database,
                quoted=self.compiler.quoted,
            )
            table = sg.table(name, catalog=database, quoted=self.compiler.quoted)
        else:
            created_table = table = sg.table(
                name, catalog=database, quoted=self.compiler.quoted
            )

        create_stmt = self._generate_create_table(
            created_table, schema=(schema or obj.schema())
        ).sql(self.name)

        with self.begin() as cur:
            cur.execute(create_stmt)

            if insert_query is not None:
                cur.execute(
                    sge.Insert(this=created_table, expression=insert_query).sql(
                        self.name
                    )
                )

            if overwrite:
                cur.execute(
                    sge.Drop(kind="TABLE", this=table, exists=True).sql(self.name)
                )
                # SQLite's ALTER TABLE statement doesn't support using a
                # fully-qualified table reference after RENAME TO. Since we
                # never rename between databases, we only need the table name
                # here.
                quoted_name = _quote(name)
                cur.execute(
                    f"ALTER TABLE {created_table.sql(self.name)} RENAME TO {quoted_name}"
                )

        if schema is None:
            return self.table(name, database=database)

        # preserve the input schema if it was provided
        return ops.DatabaseTable(
            name, schema=schema, source=self, namespace=ops.Namespace(database=database)
        ).to_expr()

    def drop_table(
        self,
        name: str,
        database: str | None = None,
        force: bool = False,
    ) -> None:
        drop_stmt = sg.exp.Drop(
            kind="TABLE",
            this=sg.table(name, catalog=database, quoted=self.compiler.quoted),
            exists=force,
        )
        with self._safe_raw_sql(drop_stmt):
            pass

    def create_view(
        self,
        name: str,
        obj: ir.Table,
        *,
        database: str | None = None,
        schema: str | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        view = sg.table(name, catalog=database, quoted=self.compiler.quoted)

        stmts = []
        if overwrite:
            stmts.append(sge.Drop(kind="VIEW", this=view, exists=True).sql(self.name))
        stmts.append(
            sge.Create(
                this=view, kind="VIEW", replace=False, expression=self.compile(obj)
            ).sql(self.name)
        )

        self._run_pre_execute_hooks(obj)

        with self.begin() as cur:
            for stmt in stmts:
                cur.execute(stmt)

        return self.table(name, database=database)

    def insert(
        self,
        table_name: str,
        obj: pd.DataFrame | ir.Table | list | dict,
        database: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """Insert data into a table.

        Parameters
        ----------
        table_name
            The name of the table to which data needs will be inserted
        obj
            The source data or expression to insert
        database
            Name of the attached database that the table is located in.
        overwrite
            If `True` then replace existing contents of table

        Raises
        ------
        NotImplementedError
            If inserting data from a different database
        ValueError
            If the type of `obj` isn't supported

        """
        table = sg.table(table_name, catalog=database, quoted=self.compiler.quoted)
        if not isinstance(obj, ir.Expr):
            obj = ibis.memtable(obj)

        self._run_pre_execute_hooks(obj)
        expr = self._to_sqlglot(obj)
        insert_stmt = sge.Insert(this=table, expression=expr).sql(self.name)
        with self.begin() as cur:
            if overwrite:
                cur.execute(f"DELETE FROM {table.sql(self.name)}")
            cur.execute(insert_stmt)



from __future__ import annotations

import decimal

import numpy as np
import pandas as pd
import pytest

import ibis
import ibis.expr.datatypes as dt
from ibis.backends.conftest import is_older_than
from ibis.backends.pandas.tests.conftest import TestConf as tm

TIMESTAMP = "2022-03-13 06:59:10.467417"


@pytest.mark.parametrize("from_", ["plain_float64", "plain_int64"])
@pytest.mark.parametrize(
    ("to", "expected"),
    [
        ("float16", "float16"),
        ("float32", "float32"),
        ("float64", "float64"),
        ("float", "float64"),
        ("int8", "int8"),
        ("int16", "int16"),
        ("int32", "int32"),
        ("int64", "int64"),
        ("string", "object"),
    ],
)
def test_cast_numeric(t, df, from_, to, expected):
    c = t[from_].cast(to)
    result = c.execute()
    assert str(result.dtype) == expected


@pytest.mark.parametrize("from_", ["float64_as_strings", "int64_as_strings"])
@pytest.mark.parametrize(
    ("to", "expected"), [("double", "float64"), ("string", "object")]
)
def test_cast_string(t, df, from_, to, expected):
    c = t[from_].cast(to)
    result = c.execute()
    assert str(result.dtype) == expected


@pytest.mark.parametrize("from_", ["array_of_int64", "array_of_float64"])
@pytest.mark.parametrize(
    ("to", "expected"),
    [("array<double>", dt.float64), ("array<int64>", dt.int64)],
)
def test_cast_array(t, from_, to, expected):
    c = t[from_].cast(to)
    result = c.execute()

    # The Series of arrays
    assert result.dtype == np.object_

    # One of the arrays in the Series
    res = result[0]
    assert isinstance(res, list)

    for v in result:
        assert v == [dt.normalize(expected, x) for x in v]


@pytest.mark.parametrize(
    ("to", "expected"),
    [
        pytest.param(
            "string",
            "object",
            marks=pytest.mark.skipif(
                is_older_than("pandas", "2.1.0"), reason="raises a NotImplementedError"
            ),
        ),
        ("int64", "int64"),
        ("double", "float64"),
        (
            dt.Timestamp("America/Los_Angeles"),
            "datetime64[ns, America/Los_Angeles]",
        ),
        (
            "timestamp('America/Los_Angeles')",
            "datetime64[ns, America/Los_Angeles]",
        ),
    ],
)
@pytest.mark.parametrize(
    "column",
    ["plain_datetimes_naive", "plain_datetimes_ny", "plain_datetimes_utc"],
)
def test_cast_timestamp_column(t, df, column, to, expected):
    c = t[column].cast(to)
    result = c.execute()
    assert str(result.dtype) == expected


@pytest.mark.parametrize(
    ("to", "expected"),
    [
        pytest.param(
            "string",
            str,
            marks=pytest.mark.skipif(
                is_older_than("pandas", "2.1.0"), reason="raises a NotImplementedError"
            ),
        ),
        ("int64", lambda x: pd.Timestamp(x).value // int(1e9)),
        ("double", lambda x: float(pd.Timestamp(x).value // int(1e9))),
        (
            dt.Timestamp("America/Los_Angeles"),
            lambda x: x.tz_localize(tz="America/Los_Angeles"),
        ),
    ],
)
def test_cast_timestamp_scalar_naive(client, to, expected):
    literal_expr = ibis.literal(pd.Timestamp(TIMESTAMP))
    value = literal_expr.cast(to)
    result = client.execute(value)
    raw = client.execute(literal_expr)
    assert result == expected(raw)


@pytest.mark.parametrize(
    ("to", "expected"),
    [
        pytest.param(
            "string",
            str,
            marks=pytest.mark.skipif(
                is_older_than("pandas", "2.1.0"), reason="raises a NotImplementedError"
            ),
        ),
        ("int64", lambda x: pd.Timestamp(x).value // int(1e9)),
        ("double", lambda x: float(pd.Timestamp(x).value // int(1e9))),
        (
            dt.Timestamp("America/Los_Angeles"),
            lambda x: x.astimezone(tz="America/Los_Angeles"),
        ),
    ],
)
@pytest.mark.parametrize("tz", ["UTC", "America/New_York"])
def test_cast_timestamp_scalar(client, to, expected, tz):
    literal_expr = ibis.literal(pd.Timestamp(TIMESTAMP).tz_localize(tz))
    value = literal_expr.cast(to)
    result = client.execute(value)
    raw = client.execute(literal_expr)
    assert result == expected(raw)


def test_timestamp_with_timezone_is_inferred_correctly(t, df):
    assert t.plain_datetimes_naive.type().equals(dt.timestamp)
    assert t.plain_datetimes_ny.type().equals(dt.Timestamp("America/New_York"))
    assert t.plain_datetimes_utc.type().equals(dt.Timestamp("UTC"))


@pytest.mark.parametrize(
    "column",
    ["plain_datetimes_naive", "plain_datetimes_ny", "plain_datetimes_utc"],
)
def test_cast_date(t, df, column):
    expr = t[column].cast("date")
    result = expr.execute()
    expected = df[column].dt.normalize().dt.tz_localize(None).dt.date
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize("type", [dt.Decimal(9, 2), dt.Decimal(12, 3)])
def test_cast_to_decimal(t, df, type):
    expr = t.float64_as_strings.cast(type)
    result = expr.execute()
    context = decimal.Context(prec=type.precision)
    expected = df.float64_as_strings.apply(
        lambda x: context.create_decimal(x).quantize(
            decimal.Decimal(
                "{}.{}".format("0" * (type.precision - type.scale), "0" * type.scale)
            )
        )
    )
    tm.assert_series_equal(result, expected)
    assert all(
        abs(element.as_tuple().exponent) == type.scale for element in result.values
    )
    assert all(
        1 <= len(element.as_tuple().digits) <= type.precision
        for element in result.values
    )



from __future__ import annotations

import pandas.testing as tm
import pytest

pytest.importorskip("pyspark")


def test_isnull(con):
    table = con.table("null_table")
    table_pandas = table.execute()

    for col, _ in table_pandas.items():
        result = table[table[col].isnull()].execute().reset_index(drop=True)
        expected = table_pandas[table_pandas[col].isnull()].reset_index(drop=True)
        tm.assert_frame_equal(result, expected)


def test_notnull(con):
    table = con.table("null_table")
    table_pandas = table.execute()

    for col, _ in table_pandas.items():
        result = table[table[col].notnull()].execute().reset_index(drop=True)
        expected = table_pandas[table_pandas[col].notnull()].reset_index(drop=True)
        tm.assert_frame_equal(result, expected)



from __future__ import annotations

import string
from functools import partial, reduce

import sqlglot as sg
import sqlglot.expressions as sge
from public import public

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.rules as rlz
from ibis.backends.base.sqlglot.compiler import NULL, STAR, SQLGlotCompiler, paren
from ibis.backends.base.sqlglot.datatypes import PostgresType
from ibis.backends.base.sqlglot.dialects import Postgres
from ibis.backends.base.sqlglot.rewrites import rewrite_sample_as_filter


class PostgresUDFNode(ops.Value):
    shape = rlz.shape_like("args")


@public
class PostgresCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = Postgres
    type_mapper = PostgresType
    rewrites = (rewrite_sample_as_filter, *SQLGlotCompiler.rewrites)

    NAN = sge.Literal.number("'NaN'::double precision")
    POS_INF = sge.Literal.number("'Inf'::double precision")
    NEG_INF = sge.Literal.number("'-Inf'::double precision")
    UNSUPPORTED_OPERATIONS = frozenset(
        (
            ops.RowID,
            ops.TimeDelta,
            ops.ArrayFlatten,
        )
    )

    SIMPLE_OPS = {
        ops.ArrayCollect: "array_agg",
        ops.ArrayRemove: "array_remove",
        ops.BitAnd: "bit_and",
        ops.BitOr: "bit_or",
        ops.BitXor: "bit_xor",
        ops.GeoArea: "st_area",
        ops.GeoAsBinary: "st_asbinary",
        ops.GeoAsEWKB: "st_asewkb",
        ops.GeoAsEWKT: "st_asewkt",
        ops.GeoAsText: "st_astext",
        ops.GeoAzimuth: "st_azimuth",
        ops.GeoBuffer: "st_buffer",
        ops.GeoCentroid: "st_centroid",
        ops.GeoContains: "st_contains",
        ops.GeoContainsProperly: "st_contains",
        ops.GeoCoveredBy: "st_coveredby",
        ops.GeoCovers: "st_covers",
        ops.GeoCrosses: "st_crosses",
        ops.GeoDFullyWithin: "st_dfullywithin",
        ops.GeoDWithin: "st_dwithin",
        ops.GeoDifference: "st_difference",
        ops.GeoDisjoint: "st_disjoint",
        ops.GeoDistance: "st_distance",
        ops.GeoEndPoint: "st_endpoint",
        ops.GeoEnvelope: "st_envelope",
        ops.GeoEquals: "st_equals",
        ops.GeoGeometryN: "st_geometryn",
        ops.GeoGeometryType: "st_geometrytype",
        ops.GeoIntersection: "st_intersection",
        ops.GeoIntersects: "st_intersects",
        ops.GeoIsValid: "st_isvalid",
        ops.GeoLength: "st_length",
        ops.GeoLineLocatePoint: "st_linelocatepoint",
        ops.GeoLineMerge: "st_linemerge",
        ops.GeoLineSubstring: "st_linesubstring",
        ops.GeoNPoints: "st_npoints",
        ops.GeoOrderingEquals: "st_orderingequals",
        ops.GeoOverlaps: "st_overlaps",
        ops.GeoPerimeter: "st_perimeter",
        ops.GeoSRID: "st_srid",
        ops.GeoSetSRID: "st_setsrid",
        ops.GeoSimplify: "st_simplify",
        ops.GeoStartPoint: "st_startpoint",
        ops.GeoTouches: "st_touches",
        ops.GeoTransform: "st_transform",
        ops.GeoUnaryUnion: "st_union",
        ops.GeoUnion: "st_union",
        ops.GeoWithin: "st_within",
        ops.GeoX: "st_x",
        ops.GeoY: "st_y",
        ops.MapContains: "exist",
        ops.MapKeys: "akeys",
        ops.MapValues: "avals",
        ops.RegexSearch: "regexp_like",
        ops.TimeFromHMS: "make_time",
    }

    def _aggregate(self, funcname: str, *args, where):
        expr = self.f[funcname](*args)
        if where is not None:
            return sge.Filter(this=expr, expression=sge.Where(this=where))
        return expr

    def visit_Mode(self, op, *, arg, where):
        expr = self.f.mode()
        expr = sge.WithinGroup(
            this=expr,
            expression=sge.Order(expressions=[sge.Ordered(this=arg)]),
        )
        if where is not None:
            expr = sge.Filter(this=expr, expression=sge.Where(this=where))
        return expr

    def visit_ArgMinMax(self, op, *, arg, key, where, desc: bool):
        conditions = [arg.is_(sg.not_(NULL)), key.is_(sg.not_(NULL))]

        if where is not None:
            conditions.append(where)

        agg = self.agg.array_agg(
            sge.Ordered(this=sge.Order(this=arg, expressions=[key]), desc=desc),
            where=sg.and_(*conditions),
        )
        return paren(agg)[0]

    def visit_ArgMin(self, op, *, arg, key, where):
        return self.visit_ArgMinMax(op, arg=arg, key=key, where=where, desc=False)

    def visit_ArgMax(self, op, *, arg, key, where):
        return self.visit_ArgMinMax(op, arg=arg, key=key, where=where, desc=True)

    def visit_Sum(self, op, *, arg, where):
        arg = (
            self.cast(self.cast(arg, dt.int32), op.dtype)
            if op.arg.dtype.is_boolean()
            else arg
        )
        return self.agg.sum(arg, where=where)

    def visit_IsNan(self, op, *, arg):
        return arg.eq(self.cast(sge.convert("NaN"), op.arg.dtype))

    def visit_IsInf(self, op, *, arg):
        return arg.isin(self.POS_INF, self.NEG_INF)

    def visit_CountDistinctStar(self, op, *, where, arg):
        # use a tuple because postgres doesn't accept COUNT(DISTINCT a, b, c, ...)
        #
        # this turns the expression into COUNT(DISTINCT ROW(a, b, c, ...))
        row = sge.Tuple(
            expressions=list(
                map(partial(sg.column, quoted=self.quoted), op.arg.schema.keys())
            )
        )
        return self.agg.count(sge.Distinct(expressions=[row]), where=where)

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "sample":
            raise com.UnsupportedOperationError(
                f"{self.dialect} only implements `pop` correlation coefficient"
            )

        # TODO: rewrite rule?
        if (left_type := op.left.dtype).is_boolean():
            left = self.cast(left, dt.Int32(nullable=left_type.nullable))

        if (right_type := op.right.dtype).is_boolean():
            right = self.cast(right, dt.Int32(nullable=right_type.nullable))

        return self.agg.corr(left, right, where=where)

    def visit_ApproxMedian(self, op, *, arg, where):
        return self.visit_Median(op, arg=arg, where=where)

    def visit_Median(self, op, *, arg, where):
        return self.visit_Quantile(op, arg=arg, quantile=sge.convert(0.5), where=where)

    def visit_ApproxCountDistinct(self, op, *, arg, where):
        return self.agg.count(sge.Distinct(expressions=[arg]), where=where)

    def visit_Range(self, op, *, start, stop, step):
        def zero_value(dtype):
            if dtype.is_interval():
                return self.f.make_interval()
            return 0

        def interval_sign(v):
            zero = self.f.make_interval()
            return sge.Case(
                ifs=[
                    self.if_(v.eq(zero), 0),
                    self.if_(v < zero, -1),
                    self.if_(v > zero, 1),
                ],
                default=NULL,
            )

        def _sign(value, dtype):
            if dtype.is_interval():
                return interval_sign(value)
            return self.f.sign(value)

        step_dtype = op.step.dtype
        return self.if_(
            sg.and_(
                self.f.nullif(step, zero_value(step_dtype)).is_(sg.not_(NULL)),
                _sign(step, step_dtype).eq(_sign(stop - start, step_dtype)),
            ),
            self.f.array_remove(
                self.f.array(
                    sg.select(STAR).from_(self.f.generate_series(start, stop, step))
                ),
                stop,
            ),
            self.cast(self.f.array(), op.dtype),
        )

    visit_IntegerRange = visit_TimestampRange = visit_Range

    def visit_StringConcat(self, op, *, arg):
        return reduce(lambda x, y: sge.DPipe(this=x, expression=y), arg)

    def visit_ArrayConcat(self, op, *, arg):
        return reduce(self.f.array_cat, map(partial(self.cast, to=op.dtype), arg))

    def visit_ArrayContains(self, op, *, arg, other):
        arg_dtype = op.arg.dtype
        return sge.ArrayContains(
            this=self.cast(arg, arg_dtype),
            expression=self.f.array(self.cast(other, arg_dtype.value_type)),
        )

    def visit_ArrayFilter(self, op, *, arg, body, param):
        return self.f.array(
            sg.select(sg.column(param, quoted=self.quoted))
            .from_(sge.Unnest(expressions=[arg], alias=param))
            .where(body)
        )

    def visit_ArrayMap(self, op, *, arg, body, param):
        return self.f.array(
            sg.select(body).from_(sge.Unnest(expressions=[arg], alias=param))
        )

    def visit_ArrayPosition(self, op, *, arg, other):
        t = sge.Unnest(expressions=[arg], alias="value", offset=True)
        idx = sg.column("ordinality")
        value = sg.column("value")
        return self.f.coalesce(
            sg.select(idx).from_(t).where(value.eq(other)).limit(1).subquery(), 0
        )

    def visit_ArraySort(self, op, *, arg):
        return self.f.array(
            sg.select("x").from_(sge.Unnest(expressions=[arg], alias="x")).order_by("x")
        )

    def visit_ArrayRepeat(self, op, *, arg, times):
        i = sg.to_identifier("i")
        length = self.f.cardinality(arg)
        return self.f.array(
            sg.select(arg[i % length + 1]).from_(
                self.f.generate_series(0, length * times - 1).as_(i.name)
            )
        )

    def visit_ArrayDistinct(self, op, *, arg):
        return self.if_(
            arg.is_(NULL), NULL, self.f.array(sg.select(self.f.explode(arg)).distinct())
        )

    def visit_ArrayUnion(self, op, *, left, right):
        return self.f.anon.array(
            sg.union(sg.select(self.f.explode(left)), sg.select(self.f.explode(right)))
        )

    def visit_ArrayIntersect(self, op, *, left, right):
        return self.f.anon.array(
            sg.intersect(
                sg.select(self.f.explode(left)), sg.select(self.f.explode(right))
            )
        )

    def visit_Log2(self, op, *, arg):
        return self.cast(
            self.f.log(
                self.cast(sge.convert(2), dt.decimal),
                arg if op.arg.dtype.is_decimal() else self.cast(arg, dt.decimal),
            ),
            op.dtype,
        )

    def visit_Log(self, op, *, arg, base):
        if base is not None:
            if not op.base.dtype.is_decimal():
                base = self.cast(base, dt.decimal)
        else:
            base = self.cast(sge.convert(self.f.exp(1)), dt.decimal)

        if not op.arg.dtype.is_decimal():
            arg = self.cast(arg, dt.decimal)
        return self.cast(self.f.log(base, arg), op.dtype)

    def visit_StructField(self, op, *, arg, field):
        idx = op.arg.dtype.names.index(field) + 1
        # postgres doesn't have anonymous structs :(
        #
        # this works around ibis not having a way to tell sqlglot to transform
        # an exploded array(row) into the equivalent unnest(t) _ (col1, ..., colN)
        # element
        #
        # but also postgres should really support anonymous structs
        return self.cast(
            self.f.json_extract(self.f.to_jsonb(arg), sge.convert(f"f{idx:d}")),
            op.dtype,
        )

    def visit_StructColumn(self, op, *, names, values):
        return self.f.row(*map(self.cast, values, op.dtype.types))

    def visit_ToJSONArray(self, op, *, arg):
        return self.if_(
            self.f.json_typeof(arg).eq(sge.convert("array")),
            self.f.array(sg.select(STAR).from_(self.f.json_array_elements(arg))),
            NULL,
        )

    def visit_Map(self, op, *, keys, values):
        return self.f.map(self.f.array(*keys), self.f.array(*values))

    def visit_MapLength(self, op, *, arg):
        return self.f.cardinality(self.f.akeys(arg))

    def visit_MapGet(self, op, *, arg, key, default):
        return self.if_(self.f.exist(arg, key), self.f.json_extract(arg, key), default)

    def visit_MapMerge(self, op, *, left, right):
        return sge.DPipe(this=left, expression=right)

    def visit_TypeOf(self, op, *, arg):
        typ = self.cast(self.f.pg_typeof(arg), dt.string)
        return self.if_(
            typ.eq(sge.convert("unknown")),
            "null" if op.arg.dtype.is_null() else "text",
            typ,
        )

    def visit_Round(self, op, *, arg, digits):
        if digits is None:
            return self.f.round(arg)

        result = self.f.round(self.cast(arg, dt.decimal), digits)
        if op.arg.dtype.is_decimal():
            return result
        return self.cast(result, dt.float64)

    def visit_Modulus(self, op, *, left, right):
        # postgres doesn't allow modulus of double precision values, so upcast and
        # then downcast later if necessary
        if not op.dtype.is_integer():
            left = self.cast(left, dt.decimal)
            right = self.cast(right, dt.decimal)

        result = left % right
        if op.dtype.is_float64():
            return self.cast(result, dt.float64)
        else:
            return result

    def visit_RegexExtract(self, op, *, arg, pattern, index):
        pattern = self.f.concat("(", pattern, ")")
        matches = self.f.regexp_match(arg, pattern)
        return self.if_(arg.rlike(pattern), paren(matches)[index], NULL)

    def visit_FindInSet(self, op, *, needle, values):
        return self.f.coalesce(
            self.f.array_position(self.f.array(*values), needle),
            0,
        )

    def visit_StringContains(self, op, *, haystack, needle):
        return self.f.strpos(haystack, needle) > 0

    def visit_EndsWith(self, op, *, arg, end):
        return self.f.right(arg, self.f.length(end)).eq(end)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_binary():
            return self.cast("".join(map(r"\x{:0>2x}".format, value)), dt.binary)
        elif dtype.is_time():
            to_int32 = partial(self.cast, to=dt.int32)
            to_float64 = partial(self.cast, to=dt.float64)

            return self.f.make_time(
                to_int32(value.hour),
                to_int32(value.minute),
                to_float64(value.second + value.microsecond / 1e6),
            )
        elif dtype.is_json():
            return self.cast(value, dt.json)
        return None

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds
    ):
        to_int32 = partial(self.cast, to=dt.int32)
        return self.f.make_timestamp(
            to_int32(year),
            to_int32(month),
            to_int32(day),
            to_int32(hours),
            to_int32(minutes),
            self.cast(seconds, dt.float64),
        )

    def visit_DateFromYMD(self, op, *, year, month, day):
        to_int32 = partial(self.cast, to=dt.int32)
        return self.f.datefromparts(to_int32(year), to_int32(month), to_int32(day))

    def visit_TimestampBucket(self, op, *, arg, interval, offset):
        origin = self.f.make_timestamp(
            *map(partial(self.cast, to=dt.int32), (1970, 1, 1, 0, 0, 0))
        )

        if offset is not None:
            origin += offset

        return self.f.date_bin(interval, arg, origin)

    def visit_DayOfWeekIndex(self, op, *, arg):
        return self.cast(self.f.extract("dow", arg) + 6, dt.int16) % 7

    def visit_DayOfWeekName(self, op, *, arg):
        return self.f.trim(self.f.to_char(arg, "Day"), string.whitespace)

    def visit_ExtractSecond(self, op, *, arg):
        return self.cast(self.f.floor(self.f.extract("second", arg)), op.dtype)

    def visit_ExtractMillisecond(self, op, *, arg):
        return self.cast(
            self.f.floor(self.f.extract("millisecond", arg)) % 1_000, op.dtype
        )

    def visit_ExtractMicrosecond(self, op, *, arg):
        return self.f.extract("microsecond", arg) % 1_000_000

    def visit_ExtractDayOfYear(self, op, *, arg):
        return self.f.extract("doy", arg)

    def visit_ExtractWeekOfYear(self, op, *, arg):
        return self.f.extract("week", arg)

    def visit_ExtractEpochSeconds(self, op, *, arg):
        return self.f.extract("epoch", arg)

    def visit_ArrayIndex(self, op, *, arg, index):
        index = self.if_(index < 0, self.f.cardinality(arg) + index, index)
        return paren(arg)[index + 1]

    def visit_ArraySlice(self, op, *, arg, start, stop):
        neg_to_pos_index = lambda n, index: self.if_(index < 0, n + index, index)

        arg_length = self.f.cardinality(arg)

        if start is None:
            start = 0
        else:
            start = self.f.least(arg_length, neg_to_pos_index(arg_length, start))

        if stop is None:
            stop = arg_length
        else:
            stop = neg_to_pos_index(arg_length, stop)

        slice_expr = sge.Slice(this=start + 1, expression=stop)
        return paren(arg)[slice_expr]

    def visit_IntervalFromInteger(self, op, *, arg, unit):
        plural = unit.plural
        if plural == "minutes":
            plural = "mins"
            arg = self.cast(arg, dt.int32)
        elif plural == "seconds":
            plural = "secs"
            arg = self.cast(arg, dt.float64)
        elif plural == "milliseconds":
            plural = "secs"
            arg /= 1e3
        elif plural == "microseconds":
            plural = "secs"
            arg /= 1e6
        elif plural == "nanoseconds":
            plural = "secs"
            arg /= 1e9
        else:
            arg = self.cast(arg, dt.int32)

        key = sg.to_identifier(plural)

        return self.f.make_interval(sge.Kwarg(this=key, expression=arg))

    def visit_Cast(self, op, *, arg, to):
        from_ = op.arg.dtype

        if from_.is_timestamp() and to.is_integer():
            return self.f.extract("epoch", arg)
        elif from_.is_integer() and to.is_timestamp():
            arg = self.f.to_timestamp(arg)
            if (timezone := to.timezone) is not None:
                arg = self.f.timezone(timezone, arg)
            return arg
        elif from_.is_integer() and to.is_interval():
            unit = to.unit
            return self.visit_IntervalFromInteger(
                ops.IntervalFromInteger(op.arg, unit), arg=arg, unit=unit
            )
        elif from_.is_string() and to.is_binary():
            # Postgres and Python use the words "decode" and "encode" in
            # opposite ways, sweet!
            return self.f.decode(arg, "escape")
        elif from_.is_binary() and to.is_string():
            return self.f.encode(arg, "escape")

        return self.cast(arg, op.to)

    visit_TryCast = visit_Cast



from __future__ import annotations

from operator import methodcaller

import numpy.testing as npt
import pandas.testing as tm
import pyarrow as pa
import pytest
from pytest import param

import ibis
from ibis.conftest import LINUX, MACOS, SANDBOXED

gpd = pytest.importorskip("geopandas")
gtm = pytest.importorskip("geopandas.testing")
shapely = pytest.importorskip("shapely")


def test_geospatial_point(zones, zones_gdf):
    coord = zones.x_cent.point(zones.y_cent).name("coord")
    # this returns GeometryArray
    gp_coord = gpd.points_from_xy(zones_gdf.x_cent, zones_gdf.y_cent)

    npt.assert_array_equal(coord.to_pandas().values, gp_coord)


# this functions are not implemented in geopandas
@pytest.mark.parametrize(
    ("operation", "keywords"),
    [
        param("as_text", {}, id="as_text"),
        param("n_points", {}, id="n_points"),
    ],
)
def test_geospatial_unary_snapshot(operation, keywords, snapshot):
    t = ibis.table([("geom", "geometry")], name="t")
    expr = getattr(t.geom, operation)(**keywords).name("tmp")
    snapshot.assert_match(ibis.to_sql(expr), "out.sql")


def test_geospatial_dwithin(snapshot):
    t = ibis.table([("geom", "geometry")], name="t")
    expr = t.geom.d_within(t.geom, 3.0).name("tmp")

    snapshot.assert_match(ibis.to_sql(expr), "out.sql")


# geospatial unary functions that return a non-geometry series
# we can test using pd.testing (tm)
@pytest.mark.parametrize(
    ("op", "keywords", "gp_op"),
    [
        param("area", {}, "area", id="area"),
        param("is_valid", {}, "is_valid", id="is_valid"),
        param(
            "geometry_type",
            {},
            "geom_type",
            id="geometry_type",
            marks=pytest.mark.xfail(raises=pa.lib.ArrowTypeError),
        ),
    ],
)
def test_geospatial_unary_tm(op, keywords, gp_op, zones, zones_gdf):
    expr = getattr(zones.geom, op)(**keywords).name("tmp")
    gp_expr = getattr(zones_gdf.geometry, gp_op)

    tm.assert_series_equal(expr.to_pandas(), gp_expr, check_names=False)


@pytest.mark.parametrize(
    ("op", "keywords", "gp_op"),
    [
        param("x", {}, "x", id="x_coord"),
        param("y", {}, "y", id="y_coord"),
    ],
)
def test_geospatial_xy(op, keywords, gp_op, zones, zones_gdf):
    cen = zones.geom.centroid().name("centroid")
    gp_cen = zones_gdf.geometry.centroid

    expr = getattr(cen, op)(**keywords).name("tmp")
    gp_expr = getattr(gp_cen, gp_op)

    tm.assert_series_equal(expr.to_pandas(), gp_expr, check_names=False)


def test_geospatial_length(lines, lines_gdf):
    # note: ST_LENGTH returns 0 for the case of polygon
    # or multi polygon while pandas geopandas returns the perimeter.
    length = lines.geom.length().name("length")
    gp_length = lines_gdf.geometry.length

    tm.assert_series_equal(length.to_pandas(), gp_length, check_names=False)


# geospatial binary functions that return a non-geometry series
# we can test using pd.testing (tm)
@pytest.mark.parametrize(
    ("op", "gp_op"),
    [
        param("contains", "contains", id="contains"),
        param("geo_equals", "geom_equals", id="geo_eqs"),
        param("covers", "covers", id="covers"),
        param("covered_by", "covered_by", id="covered_by"),
        param("crosses", "crosses", id="crosses"),
        param("disjoint", "disjoint", id="disjoint"),
        param("distance", "distance", id="distance"),
        param("intersects", "intersects", id="intersects"),
        param("overlaps", "overlaps", id="overlaps"),
        param("touches", "touches", id="touches"),
        param("within", "within", id="within"),
    ],
)
def test_geospatial_binary_tm(op, gp_op, zones, zones_gdf):
    expr = getattr(zones.geom, op)(zones.geom).name("tmp")
    gp_func = getattr(zones_gdf.geometry, gp_op)(zones_gdf.geometry)

    tm.assert_series_equal(expr.to_pandas(), gp_func, check_names=False)


# geospatial unary functions that return a geometry series
# we can test using gpd.testing (gtm)
@pytest.mark.parametrize(
    ("op", "gp_op"),
    [
        param("centroid", "centroid", id="centroid"),
        param("envelope", "envelope", id="envelope"),
    ],
)
def test_geospatial_unary_gtm(op, gp_op, zones, zones_gdf):
    expr = getattr(zones.geom, op)().name("tmp")
    gp_expr = getattr(zones_gdf.geometry, gp_op)

    gtm.assert_geoseries_equal(expr.to_pandas(), gp_expr, check_crs=False)


# geospatial binary functions that return a geometry series
# we can test using gpd.testing (gtm)
@pytest.mark.parametrize(
    ("op", "gp_op"),
    [
        param("difference", "difference", id="difference"),
        param("intersection", "intersection", id="intersection"),
        param("union", "union", id=""),
    ],
)
def test_geospatial_binary_gtm(op, gp_op, zones, zones_gdf):
    expr = getattr(zones.geom, op)(zones.geom).name("tmp")
    gp_func = getattr(zones_gdf.geometry, gp_op)(zones_gdf.geometry)

    gtm.assert_geoseries_equal(expr.to_pandas(), gp_func, check_crs=False)


def test_geospatial_end_point(lines, lines_gdf):
    epoint = lines.geom.end_point().name("end_point")
    # geopandas does not have end_point this is a work around to get it
    gp_epoint = lines_gdf.geometry.boundary.explode(index_parts=True).xs(1, level=1)

    gtm.assert_geoseries_equal(epoint.to_pandas(), gp_epoint, check_crs=False)


def test_geospatial_start_point(lines, lines_gdf):
    spoint = lines.geom.start_point().name("start_point")
    # geopandas does not have start_point this is a work around to get it
    gp_spoint = lines_gdf.geometry.boundary.explode(index_parts=True).xs(0, level=1)

    gtm.assert_geoseries_equal(spoint.to_pandas(), gp_spoint, check_crs=False)


# this one takes a bit longer than the rest.
def test_geospatial_unary_union(zones, zones_gdf):
    unary_union = zones.geom.unary_union().name("unary_union")
    # this returns a shapely geometry object
    gp_unary_union = zones_gdf.geometry.unary_union

    # using set_precision because https://github.com/duckdb/duckdb_spatial/issues/189
    assert shapely.equals(
        shapely.set_precision(unary_union.to_pandas(), grid_size=1e-7),
        shapely.set_precision(gp_unary_union, grid_size=1e-7),
    )


def test_geospatial_buffer_point(zones, zones_gdf):
    cen = zones.geom.centroid().name("centroid")
    gp_cen = zones_gdf.geometry.centroid

    buffer = cen.buffer(100.0)
    # geopandas resolution default is 16, while duckdb is 8.
    gp_buffer = gp_cen.buffer(100.0, resolution=8)

    gtm.assert_geoseries_equal(buffer.to_pandas(), gp_buffer, check_crs=False)


def test_geospatial_buffer(zones, zones_gdf):
    buffer = zones.geom.buffer(100.0)
    # geopandas resolution default is 16, while duckdb is 8.
    gp_buffer = zones_gdf.geometry.buffer(100.0, resolution=8)

    gtm.assert_geoseries_equal(buffer.to_pandas(), gp_buffer, check_crs=False)


# using a smaller dataset for time purposes
def test_geospatial_convert(geotable, gdf):
    # geotable is fabricated but let's say the
    # data is in CRS: EPSG:2263
    # let's transform to EPSG:4326 (latitude-longitude projection)
    geo_ll = geotable.geom.convert("EPSG:2263", "EPSG:4326")

    gdf.crs = "EPSG:2263"
    gdf_ll = gdf.geometry.to_crs(crs=4326)

    gtm.assert_geoseries_equal(
        geo_ll.to_pandas(), gdf_ll, check_less_precise=True, check_crs=False
    )


def test_geospatial_flip_coordinates(geotable):
    flipped = geotable.geom.flip_coordinates()

    # flipped coords
    point = shapely.geometry.Point(40, -100)
    line_string = shapely.geometry.LineString([[0, 0], [1, 1], [1, 2], [2, 2]])
    polygon = shapely.geometry.Polygon(((0, 0), (0, 1), (1, 1), (1, 0), (0, 0)))

    d = {
        "name": ["Point", "LineString", "Polygon"],
        "geometry": [point, line_string, polygon],
    }

    flipped_gdf = gpd.GeoDataFrame(d)

    gtm.assert_geoseries_equal(
        flipped.to_pandas(), flipped_gdf.geometry, check_crs=False
    )


def test_create_table_geospatial_types(geotable, con):
    name = ibis.util.gen_name("geotable")

    # con = ibis.get_backend(geotable)

    t = con.create_table(name, geotable, temp=True)

    assert t.op().name in con.list_tables()
    assert any(map(methodcaller("is_geospatial"), t.schema().values()))


# geo literals declaration
point = ibis.literal((1, 0), type="point").name("p")
point_geom = ibis.literal((1, 0), type="point:geometry").name("p")


@pytest.mark.parametrize("expr", [point, point_geom])
def test_literal_geospatial_explicit(con, expr, snapshot):
    result = str(con.compile(expr))
    snapshot.assert_match(result, "out.sql")


# test input data with shapely geometries
shp_point_0 = shapely.Point(0, 0)
shp_point_1 = shapely.Point(1, 1)
shp_point_2 = shapely.Point(2, 2)

shp_linestring_0 = shapely.LineString([shp_point_0, shp_point_1, shp_point_2])
shp_linestring_1 = shapely.LineString([shp_point_2, shp_point_1, shp_point_0])
shp_polygon_0 = shapely.Polygon(shp_linestring_0)
shp_multilinestring_0 = shapely.MultiLineString([shp_linestring_0, shp_linestring_1])
shp_multipoint_0 = shapely.MultiPoint([shp_point_0, shp_point_1, shp_point_2])
shp_multipolygon_0 = shapely.MultiPolygon([shp_polygon_0])


@pytest.mark.parametrize(
    ("shp", "expected"),
    [
        (shp_point_0, "(0 0)"),
        (shp_point_1, "(1 1)"),
        (shp_point_2, "(2 2)"),
        (shp_linestring_0, "(0 0, 1 1, 2 2)"),
        (shp_linestring_1, "(2 2, 1 1, 0 0)"),
        (shp_polygon_0, "((0 0, 1 1, 2 2, 0 0))"),
        (shp_multipolygon_0, "(((0 0, 1 1, 2 2, 0 0)))"),
        (shp_multilinestring_0, "((0 0, 1 1, 2 2), (2 2, 1 1, 0 0))"),
        param(
            shp_multipoint_0,
            "((0 0), (1 1), (2 2))",
            marks=pytest.mark.xfail(
                raises=AssertionError,
                reason="Bug-fix change in GEOS 3.12 see shapely issue #1992",
            ),
        ),
    ],
)
def test_literal_geospatial_inferred(con, shp, expected, snapshot):
    result = str(con.compile(ibis.literal(shp).name("result")))
    name = type(shp).__name__.upper()
    pair = f"{name} {expected}"
    assert pair in result
    snapshot.assert_match(result, "out.sql")


@pytest.mark.skipif(
    (LINUX or MACOS) and SANDBOXED,
    reason="nix on linux cannot download duckdb extensions or data due to sandboxing",
)
def test_load_geo_example(con):
    t = ibis.examples.zones.fetch(backend=con)
    assert t.geom.type().is_geospatial()



from __future__ import annotations

import ibis
import ibis.expr.operations as ops
from ibis.expr.rewrites import simplify

t = ibis.table(
    name="t",
    schema={
        "bool_col": "boolean",
        "int_col": "int64",
        "float_col": "float64",
        "string_col": "string",
    },
)


def test_simplify_full_reprojection():
    t1 = t.select(t)
    t1_opt = simplify(t1.op())
    assert t1_opt == t.op()


def test_simplify_subsequent_field_selections():
    t1 = t.select(t.bool_col, t.int_col, t.float_col)
    assert t1.op() == ops.Project(
        parent=t,
        values={
            "bool_col": t.bool_col,
            "int_col": t.int_col,
            "float_col": t.float_col,
        },
    )

    t2 = t1.select(t1.bool_col, t1.int_col)
    t2_opt = simplify(t2.op())
    assert t2_opt == ops.Project(
        parent=t,
        values={
            "bool_col": t.bool_col,
            "int_col": t.int_col,
        },
    )

    t3 = t2.select(t2.bool_col)
    t3_opt = simplify(t3.op())
    assert t3_opt == ops.Project(parent=t, values={"bool_col": t.bool_col})


def test_simplify_subsequent_value_selections():
    t1 = t.select(
        bool_col=~t.bool_col, int_col=t.int_col + 1, float_col=t.float_col * 3
    )
    t2 = t1.select(t1.bool_col, t1.int_col, t1.float_col)
    t2_opt = simplify(t2.op())
    assert t2_opt == ops.Project(
        parent=t,
        values={
            "bool_col": ~t.bool_col,
            "int_col": t.int_col + 1,
            "float_col": t.float_col * 3,
        },
    )

    t3 = t2.select(
        t2.bool_col,
        t2.int_col,
        float_col=t2.float_col * 2,
        another_col=t1.float_col - 1,
    )
    t3_opt = simplify(t3.op())
    assert t3_opt == ops.Project(
        parent=t,
        values={
            "bool_col": ~t.bool_col,
            "int_col": t.int_col + 1,
            "float_col": (t.float_col * 3) * 2,
            "another_col": (t.float_col * 3) - 1,
        },
    )


def test_simplify_subsequent_filters():
    f1 = t.filter(t.bool_col)
    f2 = f1.filter(t.int_col > 0)
    f2_opt = simplify(f2.op())
    assert f2_opt == ops.Filter(t, predicates=[t.bool_col, t.int_col > 0])


def test_simplify_project_filter_project():
    t1 = t.select(
        bool_col=~t.bool_col, int_col=t.int_col + 1, float_col=t.float_col * 3
    )
    t2 = t1.filter(t1.bool_col)
    t3 = t2.filter(t2.int_col > 0)
    t4 = t3.select(t3.bool_col, t3.int_col)

    filt = ops.Filter(parent=t, predicates=[~t.bool_col, t.int_col + 1 > 0]).to_expr()
    proj = ops.Project(
        parent=filt, values={"bool_col": ~filt.bool_col, "int_col": filt.int_col + 1}
    ).to_expr()

    t4_opt = simplify(t4.op())
    assert t4_opt == proj.op()
